#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Generate Glances Python API documentation."""

from pprint import pformat

from glances import api

APIDOC_HEADER = """\
.. _api:

Python API documentation
========================

This documentation describes the Glances Python API.

Note: This API is only available in Glances 4.4.0 or higher.

"""


def printtab(s, indent='    '):
    print(indent + s.replace('\n', '\n' + indent))


def print_tldr(gl):
    """Print the TL;DR section of the API documentation."""
    sub_title = 'TL;DR'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('You can access the Glances API by importing the `glances.api` module and creating an')
    print('instance of the `GlancesAPI` class. This instance provides access to all Glances plugins')
    print('and their fields. For example, to access the CPU plugin and its total field, you can')
    print('use the following code:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> from glances import api')
    printtab('>>> gl = api.GlancesAPI()')
    printtab('>>> gl.cpu')
    printtab(f'{pformat(gl.cpu.stats)}')
    printtab('>>> gl.cpu.get("total")')
    printtab(f'{gl.cpu.get("total")}')
    printtab('>>> gl.mem.get("used")')
    printtab(f'{gl.mem.get("used")}')
    printtab('>>> gl.auto_unit(gl.mem.get("used"))')
    printtab(f'{gl.auto_unit(gl.mem.get("used"))}')
    print('')
    print('If the stats return a list of items (like network interfaces or processes), you can')
    print('access them by their name:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> gl.network.keys()')
    printtab(f'{gl.network.keys()}')
    printtab(f'>>> gl.network["{gl.network.keys()[0]}"]')
    printtab(f'{pformat(gl.network[gl.network.keys()[0]])}')
    print('')


def print_init_api(gl):
    sub_title = 'Init Glances Python API'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Init the Glances API:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> from glances import api')
    printtab('>>> gl = api.GlancesAPI()')
    print('')


def print_plugins_list(gl):
    sub_title = 'Get Glances plugins list'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Get the plugins list:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> gl.plugins()')
    printtab(f'{gl.plugins()}')
    print('')


def print_plugin(gl, plugin):
    """Print the details of a single plugin."""
    sub_title = f'Glances {plugin}'
    print(sub_title)
    print('-' * len(sub_title))
    print('')

    stats_obj = gl.__getattr__(plugin)

    print(f'{plugin.capitalize()} stats:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab(f'>>> type(gl.{plugin})')
    printtab(f'{type(stats_obj)}')
    if len(stats_obj.keys()) > 0 and isinstance(stats_obj[stats_obj.keys()[0]], dict):
        printtab(f'>>> gl.{plugin}')
        printtab(f'Return a dict of dict with key=<{stats_obj[stats_obj.keys()[0]]["key"]}>')
        printtab(f'>>> gl.{plugin}.keys()')
        printtab(f'{stats_obj.keys()}')
        printtab(f'>>> gl.{plugin}.get("{stats_obj.keys()[0]}")')
        printtab(f'{pformat(stats_obj[stats_obj.keys()[0]])}')
    else:
        printtab(f'>>> gl.{plugin}')
        printtab(f'{pformat(stats_obj.stats)}')
        if len(stats_obj.keys()) > 0:
            printtab(f'>>> gl.{plugin}.keys()')
            printtab(f'{stats_obj.keys()}')
            printtab(f'>>> gl.{plugin}.get("{stats_obj.keys()[0]}")')
            printtab(f'{pformat(stats_obj[stats_obj.keys()[0]])}')
    print('')

    if stats_obj.fields_description is not None:
        print(f'{plugin.capitalize()} fields description:')
        print('')
        for field, description in stats_obj.fields_description.items():
            print(f'* {field}: {description["description"]}')
        print('')

    print(f'{plugin.capitalize()} limits:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab(f'>>> gl.{plugin}.limits')
    printtab(f'{pformat(gl.__getattr__(plugin).limits)}')
    print('')


def print_plugins(gl):
    """Print the details of all plugins."""
    for plugin in [p for p in gl.plugins() if p not in ['help', 'programlist']]:
        print_plugin(gl, plugin)


def print_auto_unit(gl):
    sub_title = 'Use auto_unit to display a human-readable string with the unit'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Use auto_unit() function to generate a human-readable string with the unit:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> gl.mem.get("used")')
    printtab(f'{gl.mem.get("used")}')
    print('')
    printtab('>>> gl.auto_unit(gl.mem.get("used"))')
    printtab(f'{gl.auto_unit(gl.mem.get("used"))}')
    print('')
    print("""
Args:

    number (float or int): The numeric value to be converted.

    low_precision (bool, optional): If True, use lower precision for the output. Defaults to False.

    min_symbol (str, optional): The minimum unit symbol to use (e.g., 'K' for kilo). Defaults to 'K'.

    none_symbol (str, optional): The symbol to display if the number is None. Defaults to '-'.

Returns:

    str: A human-readable string representation of the number with units.

""")


def print_bar(gl):
    sub_title = 'Use to display stat as a bar'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Use bar() function to generate a bar:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> gl.bar(gl.mem["percent"])')
    printtab(f'{gl.bar(gl.mem.get("percent"))}')
    print('')
    print("""
Args:

    value (float): The percentage value to represent in the bar (typically between 0 and 100).

    size (int, optional): The total length of the bar in characters. Defaults to 18.

    bar_char (str, optional): The character used to represent the filled portion of the bar. Defaults to '■'.

    empty_char (str, optional): The character used to represent the empty portion of the bar. Defaults to '□'.

    pre_char (str, optional): A string to prepend to the bar. Defaults to ''.

    post_char (str, optional): A string to append to the bar. Defaults to ''.

Returns:

    str: A string representing the progress bar.

""")


def print_top_process(gl):
    sub_title = 'Use to display top process list'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Use top_process() function to generate a list of top processes sorted by CPU or MEM usage:')
    print('')
    print('.. code-block:: python')
    print('')
    printtab('>>> gl.top_process()')
    printtab(f'{gl.top_process()}')
    print('')
    print("""
Args:

    limit (int, optional): The maximum number of top processes to return. Defaults to 3.

    sorted_by (str, optional): The primary key to sort processes by (e.g., 'cpu_percent').
                                Defaults to 'cpu_percent'.

    sorted_by_secondary (str, optional): The secondary key to sort processes by if primary keys are equal
                                            (e.g., 'memory_percent'). Defaults to 'memory_percent'.

Returns:

    list: A list of dictionaries representing the top processes, excluding those with 'glances' in their
            command line.

Note:

    The 'glances' process is excluded from the returned list to avoid self-generated CPU load affecting
    the results.

""")


class GlancesStdoutApiDoc:
    """This class manages the fields description display."""

    def __init__(self, config=None, args=None):
        # Init
        self.gl = api.GlancesAPI()

    def end(self):
        pass

    def update(self, stats, duration=1):
        """Display issue"""
        # Display header
        print(APIDOC_HEADER)

        # Display TL;DR section
        print_tldr(self.gl)

        # Init the API
        print_init_api(self.gl)

        # Display plugins list
        print_plugins_list(self.gl)

        # Loop over plugins
        print_plugins(self.gl)

        # Others helpers
        print_auto_unit(self.gl)
        print_bar(self.gl)
        print_top_process(self.gl)

        # Return True to exit directly (no refresh)
        return True



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""AMD Extension unit for Glances' GPU plugin.

The class grabs the stats from the /sys/class/drm/ directory.

See: https://wiki.archlinux.org/title/AMDGPU#Manually
"""

# Example
# tests-data/plugins/gpu/amd/
# └── sys
#     ├── class
#     │   └── drm
#     │       └── card0
#     │           └── device
#     │               ├── device
#     │               ├── gpu_busy_percent
#     │               ├── hwmon
#     │               │   └── hwmon0
#     │               │       ├── in1_input
#     │               │       └── temp1_input
#     │               ├── mem_info_gtt_total
#     │               ├── mem_info_gtt_used
#     │               ├── mem_info_vram_total
#     │               ├── mem_info_vram_used
#     │               ├── pp_dpm_mclk
#     │               ├── pp_dpm_sclk
#     │               └── revision
#     └── kernel
#         └── debug
#             └── dri
#                 └── 0
#                     └── amdgpu_pm_info

import functools
import glob
import os
import re

DRM_ROOT_FOLDER: str = '/sys/class/drm'
DEVICE_FOLDER_PATTERN: str = 'card[0-9]/device'
AMDGPU_IDS_FILE: str = '/usr/share/libdrm/amdgpu.ids'
PCI_DEVICE_ID: str = 'device'
PCI_REVISION_ID: str = 'revision'
GPU_PROC_PERCENT: str = 'gpu_busy_percent'
GPU_MEM_TOTAL: str = 'mem_info_vram_total'
GPU_MEM_USED: str = 'mem_info_vram_used'
GTT_MEM_TOTAL: str = 'mem_info_gtt_total'
GTT_MEM_USED: str = 'mem_info_gtt_used'
HWMON_NORTHBRIDGE_VOLTAGE_PATTERN: str = 'hwmon/hwmon[0-9]/in1_input'
HWMON_TEMPERATURE_PATTERN = 'hwmon/hwmon[0-9]/temp[0-9]_input'


class AmdGPU:
    """GPU card class."""

    def __init__(self, drm_root_folder: str = DRM_ROOT_FOLDER):
        """Init AMD  GPU card class."""
        self.drm_root_folder = drm_root_folder
        self.device_folders = get_device_list(drm_root_folder)

    def exit(self):
        """Close AMD GPU class."""

    def get_device_stats(self):
        """Get AMD GPU stats."""
        stats = []

        for index, device in enumerate(self.device_folders):
            device_stats = {}
            # Dictionary key is the GPU_ID
            device_stats['key'] = 'gpu_id'
            # GPU id (for multiple GPU, start at 0)
            device_stats['gpu_id'] = f'amd{index}'
            # GPU name
            device_stats['name'] = get_device_name(device)
            # Memory consumption in % (not available on all GPU)
            device_stats['mem'] = get_mem(device)
            # Processor consumption in %
            device_stats['proc'] = get_proc(device)
            # Processor temperature in °C
            device_stats['temperature'] = get_temperature(device)
            # Fan speed in %
            device_stats['fan_speed'] = get_fan_speed(device)
            stats.append(device_stats)

        return stats


def get_device_list(drm_root_folder: str) -> list[str]:
    """Return a list of path to the device stats."""
    ret = []
    for device_folder in glob.glob(DEVICE_FOLDER_PATTERN, root_dir=drm_root_folder):
        if os.path.isfile(os.path.join(drm_root_folder, device_folder, GPU_PROC_PERCENT)):
            # If the GPU busy file is present then take the card into account
            ret.append(os.path.join(drm_root_folder, device_folder))
    return ret


def read_file(*path_segments: str) -> str | None:
    """Return content of file or None if not accessible."""
    path = os.path.join(*path_segments)
    if os.path.isfile(path):
        try:
            with open(path) as f:
                return f.read().strip()
        except (PermissionError, OSError):
            # File exists but is not readable (e.g. Snap strict confinement)
            # Graceful degradation: caller will use a default value instead
            return None
    return None


@functools.cache
def get_device_name(device_folder: str) -> str:
    """Return the GPU name."""

    # Table source: https://cgit.freedesktop.org/drm/libdrm/tree/data/amdgpu.ids
    device_id = read_file(device_folder, PCI_DEVICE_ID)
    revision_id = read_file(device_folder, PCI_REVISION_ID)
    amdgpu_ids = read_file(AMDGPU_IDS_FILE)
    if device_id and revision_id and amdgpu_ids:
        # Strip leading "0x" and convert to uppercase hexadecimal
        device_id = device_id[2:].upper()
        revision_id = revision_id[2:].upper()
        # Syntax:
        # device_id,	revision_id,	product_name        <-- single tab after comma
        pattern = re.compile(f'^{device_id},\\s{revision_id},\\s(?P<product_name>.+)$', re.MULTILINE)
        if match := pattern.search(amdgpu_ids):
            return match.group('product_name').removeprefix('AMD ').removesuffix(' Graphics')

    return 'AMD GPU'


def get_mem(device_folder: str) -> int | None:
    """Return the memory consumption in %."""
    mem_info_total = read_file(device_folder, GPU_MEM_TOTAL)
    mem_info_used = read_file(device_folder, GPU_MEM_USED)
    if mem_info_total and mem_info_used:
        mem_info_total = int(mem_info_total)
        mem_info_used = int(mem_info_used)
        # Detect integrated GPU by looking for APU-only Northbridge voltage.
        # See https://docs.kernel.org/gpu/amdgpu/thermal.html
        if glob.glob(HWMON_NORTHBRIDGE_VOLTAGE_PATTERN, root_dir=device_folder):
            mem_info_gtt_total = read_file(device_folder, GTT_MEM_TOTAL)
            mem_info_gtt_used = read_file(device_folder, GTT_MEM_USED)
            if mem_info_gtt_total and mem_info_gtt_used:
                # Integrated GPU allocates static VRAM and dynamic GTT from the same system memory.
                mem_info_total += int(mem_info_gtt_total)
                mem_info_used += int(mem_info_gtt_used)
        if mem_info_total > 0:
            return round(mem_info_used / mem_info_total * 100)
    return None


def get_proc(device_folder: str) -> int | None:
    """Return the processor consumption in %."""
    if gpu_busy_percent := read_file(device_folder, GPU_PROC_PERCENT):
        return int(gpu_busy_percent)
    return None


def get_temperature(device_folder: str) -> int | None:
    """Return the processor temperature in °C (mean of all HWMON)"""
    temp_input = []
    for temp_file in glob.glob(HWMON_TEMPERATURE_PATTERN, root_dir=device_folder):
        if a_temp_input := read_file(device_folder, temp_file):
            temp_input.append(int(a_temp_input))
        else:
            return None
    if temp_input:
        return round(sum(temp_input) / len(temp_input) / 1000)
    return None


def get_fan_speed(device_folder: str) -> int | None:
    """Return the fan speed in %."""
    return None



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""The stats server manager."""

import collections

from glances.logger import logger
from glances.stats import GlancesStats


class GlancesStatsServer(GlancesStats):
    """This class stores, updates and gives stats for the server."""

    def __init__(self, config=None, args=None):
        # Init the stats
        super().__init__(config=config, args=args)

        # Init the all_stats dict used by the server
        # all_stats is a dict of dicts filled by the server
        self.all_stats = collections.defaultdict(dict)

        # In the update method, disable extended process stats
        logger.info("Disable extended processes stats in server mode")

    def update(self, input_stats=None):
        """Update the stats."""
        input_stats = input_stats or {}

        # Force update of all the stats
        super().update()

        # Disable the extended processes stats because it cause an high CPU load
        self._plugins['processcount'].disable_extended()

        # Build all_stats variable (concatenation of all the stats)
        self.all_stats = self._set_stats(input_stats)

    def _set_stats(self, input_stats):
        """Set the stats to the input_stats one."""
        # Build the all_stats with the get_raw() method of the plugins
        return {p: self._plugins[p].get_raw() for p in self._plugins if self._plugins[p].is_enabled()}

    def getAll(self):
        """Return the stats as a list."""
        return self.all_stats



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Statsd interface class."""

from numbers import Number

from statsd import StatsClient

from glances.exports.export import GlancesExport
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the Statsd export module."""

    def __init__(self, config=None, args=None):
        """Init the Statsd export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        # N/A

        # Optional configuration keys
        self.prefix = None

        # Load the configuration file
        self.export_enable = self.load_conf('statsd', mandatories=['host', 'port'], options=['prefix'])
        if not self.export_enable:
            exit('Missing STATSD config')

        # Default prefix for stats is 'glances'
        if self.prefix is None:
            self.prefix = 'glances'

        # Init the Statsd client
        self.client = self.init()

    def init(self):
        """Init the connection to the Statsd server."""
        if not self.export_enable:
            return None
        logger.info(f"Stats will be exported to StatsD server: {self.host}:{self.port}")
        return StatsClient(self.host, int(self.port), prefix=self.prefix)

    def export(self, name, columns, points):
        """Export the stats to the Statsd server."""
        for i in range(len(columns)):
            if not isinstance(points[i], Number):
                continue
            stat_name = f'{name}.{columns[i]}'
            stat_value = points[i]
            try:
                self.client.gauge(normalize(stat_name), stat_value)
            except Exception as e:
                logger.error(f"Can not export stats to Statsd ({e})")
        logger.debug(f"Export {name} stats to Statsd")


def normalize(name):
    """Normalize name for the Statsd convention"""

    # Name should not contain some specials chars (issue #1068)
    ret = name.replace(':', '')
    ret = ret.replace('%', '')
    return ret.replace(' ', '_')



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Graph exporter interface class."""

import errno
import os
import sys
import tempfile

import pygal.style
from pygal import DateTimeLine

from glances.exports.export import GlancesExport
from glances.globals import time_series_subsample
from glances.logger import logger
from glances.timer import Timer


class Export(GlancesExport):
    """This class manages the Graph export module."""

    def __init__(self, config=None, args=None):
        """Init the export IF."""
        super().__init__(config=config, args=args)

        # Load the Graph configuration file section (is exists)
        self.export_enable = self.load_conf('graph', options=['path', 'generate_every', 'width', 'height', 'style'])

        # Manage options (command line arguments overwrite configuration file)
        self.path = self.path or args.export_graph_path
        self.generate_every = int(getattr(self, 'generate_every', 0) or 0)
        self.width = int(getattr(self, 'width', 800) or 800)
        self.height = int(getattr(self, 'height', 600) or 600)
        self.style = (
            getattr(pygal.style, getattr(self, 'style', 'DarkStyle'), pygal.style.DarkStyle) or pygal.style.DarkStyle
        )

        # Create export folder
        try:
            os.makedirs(self.path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                logger.critical(f"Cannot create the Graph output folder {self.path} ({e})")
                sys.exit(2)

        # Check if output folder is writeable
        try:
            tempfile.TemporaryFile(dir=self.path)
        except OSError:
            logger.critical(f"Graph output folder {self.path} is not writeable")
            sys.exit(2)

        logger.info(f"Graphs will be created in the {self.path} folder")
        if self.generate_every != 0:
            logger.info(f"Graphs will be created automatically every {self.generate_every} seconds")
            logger.info("or when 'g' key is pressed (only through the CLI interface)")
            # Start the timer
            self._timer = Timer(self.generate_every)
        else:
            logger.info("Graphs will be created  when 'g' key is pressed (in the CLI interface)")
            self._timer = None

    def exit(self):
        """Close the files."""
        logger.debug(f"Finalise export interface {self.export_name}")

    def update(self, stats):
        """Generate Graph file in the output folder."""

        if self.generate_every != 0 and self._timer.finished():
            self.args.generate_graph = True
            self._timer.reset()

        if not self.args.generate_graph:
            return

        plugins = stats.getPluginsList()
        for plugin_name in plugins:
            plugin = stats._plugins[plugin_name]
            if plugin_name in self.plugins_to_export(stats):
                self.export(plugin_name, plugin.get_export_history())

        logger.info(f"Graphs created in {self.path}")
        self.args.generate_graph = False

    def export(self, title, data):
        """Generate graph from the data.

        Example for the mem plugin:
        {'percent': [
            (datetime.datetime(2018, 3, 24, 16, 27, 47, 282070), 51.8),
            (datetime.datetime(2018, 3, 24, 16, 27, 47, 540999), 51.9),
            (datetime.datetime(2018, 3, 24, 16, 27, 50, 653390), 52.0),
            (datetime.datetime(2018, 3, 24, 16, 27, 53, 749702), 52.0),
            (datetime.datetime(2018, 3, 24, 16, 27, 56, 825660), 52.0),
            ...
            ]
        }

        Return:
        * True if the graph have been generated
        * False if the graph have not been generated
        """
        if data == {}:
            return False

        chart = DateTimeLine(
            title=title.capitalize(),
            width=self.width,
            height=self.height,
            style=self.style,
            show_dots=False,
            legend_at_bottom=True,
            x_label_rotation=20,
            x_value_formatter=lambda dt: dt.strftime('%Y/%m/%d %H:%M:%S'),
        )
        for k, v in time_series_subsample(data, self.width).items():
            chart.add(k, v)
        chart.render_to_file(os.path.join(self.path, title + '.svg'))
        return True



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""IP plugin."""

import threading

from glances.globals import get_ip_address, json_loads, urlopen_auth
from glances.logger import logger
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: is it a rate ? If yes, // by time_since_update when displayed,
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'address': {
        'description': 'Private IP address',
    },
    'mask': {
        'description': 'Private IP mask',
    },
    'mask_cidr': {
        'description': 'Private IP mask in CIDR format',
        'unit': 'number',
    },
    'gateway': {
        'description': 'Private IP gateway',
    },
    'public_address': {
        'description': 'Public IP address',
    },
    'public_info_human': {
        'description': 'Public IP information',
    },
}


class IpPlugin(GlancesPluginModel):
    """Glances IP Plugin.

    stats is a dict
    """

    _default_public_refresh_interval = 300

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(args=args, config=config, fields_description=fields_description)

        # We want to display the stat in the curse interface
        self.display_curse = True

        # Public information (see issue #2732)
        self.public_address = ""
        self.public_info = ""
        self.public_api = self.get_conf_value("public_api", default=[None])[0]
        self.public_username = self.get_conf_value("public_username", default=[None])[0]
        self.public_password = self.get_conf_value("public_password", default=[None])[0]
        self.public_field = self.get_conf_value("public_field", default=[None])
        self.public_template = self.get_conf_value("public_template", default=[None])[0]
        self.public_disabled = (
            self.get_conf_value('public_disabled', default='False')[0].lower() != 'false'
            or self.public_api is None
            or self.public_field is None
        )
        self.public_address_refresh_interval = self.get_conf_value(
            "public_refresh_interval", default=self._default_public_refresh_interval
        )

        # Init thread to grab public IP address asynchronously
        self.public_ip_thread = None
        if not self.public_disabled:
            self.public_ip_thread = ThreadPublicIpAddress(
                url=self.public_api,
                username=self.public_username,
                password=self.public_password,
                refresh_interval=self.public_address_refresh_interval,
            )
            self.public_ip_thread.start()

    def get_first_ip(self, stats):
        stats['address'], stats['mask'] = get_ip_address()
        stats['mask_cidr'] = self.ip_to_cidr(stats['mask'])

        return stats

    def get_public_ip(self, stats):
        """Get public IP information from the background thread (non-blocking)."""
        if self.public_ip_thread is None:
            return stats

        try:
            # Read public IP info from the background thread (non-blocking)
            self.public_info = self.public_ip_thread.public_info
            if self.public_info:
                self.public_address = self.public_info.get('ip', '')
        except (KeyError, AttributeError, TypeError) as e:
            logger.debug(f"Cannot grab public IP information ({e})")

        if self.public_address:
            stats['public_address'] = (
                self.public_address if not self.args.hide_public_info else self.__hide_ip(self.public_address)
            )
            stats['public_info_human'] = self.public_info_for_human(self.public_info)

        return stats

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update IP stats using the input method.

        :return: the stats dict
        """
        # Init new stats
        stats = self.get_init_value()

        if self.input_method == 'local':
            stats = self.get_stats_for_local_input(stats)

        elif self.input_method == 'snmp':
            # Not implemented yet
            pass

        # Update the stats
        self.stats = stats

        return self.stats

    def get_stats_for_local_input(self, stats):
        # Get Public and Private IP address
        if self.get_first_ip(stats):
            self.get_public_ip(stats)
        return stats

    def exit(self):
        """Overwrite the exit method to close the thread."""
        if self.public_ip_thread is not None:
            self.public_ip_thread.stop()
        # Call the parent class
        super().exit()

    def __hide_ip(self, ip):
        """Hide last to digit of the given IP address"""
        return '.'.join(ip.split('.')[0:2]) + '.*.*'

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        # Init the return message
        ret = []

        # Only process if stats exist and display plugin enable...
        if not self.stats or self.is_disabled():
            return ret

        # Start with the private IP information
        if 'address' in self.stats:
            msg = 'IP '
            ret.append(self.curse_add_line(msg, 'TITLE', optional=True))
            msg = '{}'.format(self.stats['address'])
            ret.append(self.curse_add_line(msg, optional=True))
        if 'mask_cidr' in self.stats:
            # VPN with no internet access (issue #842)
            msg = '/{}'.format(self.stats['mask_cidr'])
            ret.append(self.curse_add_line(msg, optional=True))

        # Then with the public IP information
        try:
            msg_pub = '{}'.format(self.stats['public_address'])
        except (UnicodeEncodeError, KeyError):
            # Add KeyError exception (see https://github.com/nicolargo/glances/issues/1469)
            pass
        else:
            if self.stats['public_address']:
                msg = ' Pub '
                ret.append(self.curse_add_line(msg, 'TITLE', optional=True))
                ret.append(self.curse_add_line(msg_pub, optional=True))

            if self.stats['public_info_human']:
                ret.append(self.curse_add_line(' {}'.format(self.stats['public_info_human']), optional=True))

        return ret

    def public_info_for_human(self, public_info):
        """Return the data to pack to the client."""
        if not public_info:
            return ''

        return self.public_template.format(**public_info)

    @staticmethod
    def ip_to_cidr(ip):
        """Convert IP address to CIDR.

        Example: '255.255.255.0' will return 24
        """
        # Thanks to @Atticfire
        # See https://github.com/nicolargo/glances/issues/1417#issuecomment-469894399
        if ip is None:
            # Correct issue #1528
            return 0
        return sum(bin(int(x)).count('1') for x in ip.split('.'))


class ThreadPublicIpAddress(threading.Thread):
    """Thread class to fetch public IP address asynchronously.

    This prevents blocking the main Glances startup and update cycle.
    """

    def __init__(self, url, username, password, refresh_interval, timeout=2):
        """Init the thread.

        :param url: URL of the public IP API
        :param username: Optional username for API authentication
        :param password: Optional password for API authentication
        :param refresh_interval: Time in seconds between refreshes
        :param timeout: Timeout for the API request
        """
        logger.debug("IP plugin - Create thread for public IP address")
        super().__init__()
        self.daemon = True
        self._stopper = threading.Event()

        self.url = url
        self.username = username
        self.password = password
        self.refresh_interval = refresh_interval
        self.timeout = timeout

        # Public IP information (shared with main thread)
        self._public_info = {}
        self._public_info_lock = threading.Lock()

    def run(self):
        """Grab public IP information in a loop.

        Runs until stop() is called, refreshing at the configured interval.
        """
        while not self._stopper.is_set():
            # Fetch public IP information
            info = self._fetch_public_ip_info()
            if info is not None:
                with self._public_info_lock:
                    self._public_info = info

            # Wait for the refresh interval or until stopped
            self._stopper.wait(self.refresh_interval)

    def _fetch_public_ip_info(self):
        """Fetch public IP information from the configured API."""
        try:
            response = urlopen_auth(self.url, self.username, self.password, self.timeout).read()
            return json_loads(response)
        except Exception as e:
            logger.debug(f"IP plugin - Cannot get public IP information from {self.url} ({e})")
            return None

    @property
    def public_info(self):
        """Return the current public IP information (thread-safe)."""
        with self._public_info_lock:
            return self._public_info.copy()

    def stop(self, timeout=None):
        """Stop the thread."""
        logger.debug("IP plugin - Close thread for public IP address")
        self._stopper.set()

    def stopped(self):
        """Return True if the thread is stopped."""
        return self._stopper.is_set()



#
# This file is part of Glances.
#
# Copyright (C) 2026 Nicolas Hennion <nicolashennion@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""NPU plugin for Glances.

Currently supported:
- AMD NPU (Ryzen AI - Phoenix, Hawk Point, Strix Point)
- Intel NPU (Meteor Lake, Lunar Lake, Arrow Lake)
- Rockchip NPU (RK3588, RK3576)
"""

from glances.logger import logger
from glances.plugins.npu.cards.amd import AmdNPU
from glances.plugins.npu.cards.intel import IntelNPU
from glances.plugins.npu.cards.rockchip import RockchipNPU
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: is it a rate ? If yes, // by time_since_update when displayed,
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'npu_id': {
        'description': 'NPU identification',
    },
    'name': {
        'description': 'NPU name',
    },
    'load': {
        'description': 'NPU load',
        'unit': 'percent',
    },
    'freq': {
        'description': 'NPU frequency',
        'unit': 'percent',
    },
    'freq_current': {
        'description': 'NPU current frequency',
        'unit': 'hertz',
    },
    'freq_max': {
        'description': 'NPU maximum frequency',
        'unit': 'hertz',
    },
    'mem': {
        'description': 'Memory consumption',
        'unit': 'percent',
    },
    'memory_used': {
        'description': 'Memory used',
        'unit': 'byte',
    },
    'memory_total': {
        'description': 'Memory total',
        'unit': 'byte',
    },
    'temperature': {
        'description': 'NPU temperature',
        'unit': 'celsius',
    },
    'power': {
        'description': 'NPU power consumption',
        'unit': 'watt',
    },
}

# Define the history items list
# All items in this list will be historised if the --enable-history tag is set
items_history_list = [
    {'name': 'freq', 'description': 'NPU processor frequency', 'y_unit': '%'},
    {'name': 'load', 'description': 'NPU processor load', 'y_unit': '%'},
    {'name': 'mem', 'description': 'Memory consumption', 'y_unit': '%'},
]


class NpuPlugin(GlancesPluginModel):
    """Glances NPU plugin.

    stats is a list of dictionaries with one entry per NPU
    """

    def __init__(
        self,
        args=None,
        config=None,
        amd_npu_root_folder: str = '/',
        intel_npu_root_folder: str = '/',
        rockchip_npu_root_folder: str = '/',
    ):
        """Init the plugin."""
        super().__init__(
            args=args,
            config=config,
            items_history_list=items_history_list,
            stats_init_value=[],
            fields_description=fields_description,
        )

        # Init the AMD NPU
        # Just for test purpose (uncomment to test on computer without AMD NPU)
        # amd_npu_root_folder = './tests-data/plugins/npu/amd'
        self.amd = AmdNPU(npu_root_folder=amd_npu_root_folder)

        # Init the Intel NPU
        # Just for test purpose (uncomment to test on computer without Intel NPU)
        # intel_npu_root_folder = './tests-data/plugins/npu/intel'
        self.intel = IntelNPU(npu_root_folder=intel_npu_root_folder)

        # Init the Rockchip NPU
        # Just for test purpose (uncomment to test on computer without Rockchip NPU)
        # rockchip_npu_root_folder = './tests-data/plugins/npu/rockchip'
        self.rockchip = RockchipNPU(npu_root_folder=rockchip_npu_root_folder)

        # We want to display the stat in the curse interface
        self.display_curse = True

    def exit(self):
        """Overwrite the exit method to close the NPU API."""
        self.amd.exit()
        self.intel.exit()
        self.rockchip.exit()

        # Call the father exit method
        super().exit()

    def get_key(self):
        """Return the key of the list."""
        return 'npu_id'

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update the NPU stats."""
        # Init new stats
        stats = self.get_init_value()

        # Get the stats
        if self.amd.is_available():
            try:
                stats.append(self.amd.get_device_stats())
            except Exception as e:
                logger.debug(f"Error getting AMD NPU stats, disable AMD NPU: {e}")
                self.amd.disable()
        if self.intel.is_available():
            try:
                stats.append(self.intel.get_device_stats())
            except Exception as e:
                logger.debug(f"Error getting Intel NPU stats, disable Intel NPU: {e}")
                self.intel.disable()
        if self.rockchip.is_available():
            try:
                stats.append(self.rockchip.get_device_stats())
            except Exception as e:
                logger.debug(f"Error getting Rockchip NPU stats, disable Rockchip NPU: {e}")
                self.rockchip.disable()

        # Update the stats
        self.stats = stats

        return self.stats

    def update_views(self):
        """Update stats views."""
        # Call the father's method
        super().update_views()

        # Add specifics information
        # Alert
        for i in self.stats:
            # Init the views for the current GPU
            self.views[i[self.get_key()]] = {'load': {}, 'freq': {}, 'mem': {}}
            # Load alert
            if 'load' in i:
                alert = self.get_alert(i['load'], header='load')
                self.views[i[self.get_key()]]['load']['decoration'] = alert
            # Frequency alert
            if 'freq' in i:
                alert = self.get_alert(i['freq'], header='freq')
                self.views[i[self.get_key()]]['freq']['decoration'] = alert
            # Memory alert
            if 'mem' in i:
                alert = self.get_alert(i['mem'], header='mem')
                self.views[i[self.get_key()]]['mem']['decoration'] = alert

        return True

    def _format_value(self, value, unit='%'):
        """Format a value with unit, or return N/A if value is None."""
        if value is None:
            return '{:>5}'.format('N/A')
        return f'{value:>4.0f}{unit}'

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        ret = []

        # Only process if stats exist, not empty and plugin not disabled
        if not self.stats or self.is_disabled():
            return ret

        # First name is NPU name (limited to 15 characters)
        ret.append(self.curse_add_line(self.stats[0]['name'][:17], 'TITLE'))

        # Second line is load (or frequency) percent and frequency limit
        ret.append(self.curse_new_line())
        if self.stats[0]['load'] is not None:
            msg = '{:<3.0%}'.format(self.stats[0]['load'] / 100)
            ret.append(
                self.curse_add_line(
                    msg, self.get_views(key='load', item=self.stats[0][self.get_key()], option='decoration')
                )
            )
        else:
            msg = '{:<3.0%}'.format(self.stats[0]['freq'] / 100)
            ret.append(
                self.curse_add_line(
                    msg, self.get_views(key='freq', item=self.stats[0][self.get_key()], option='decoration')
                )
            )
        freq = '{}/{}Hz'.format(
            self.auto_unit(self.stats[0]['freq_current']), self.auto_unit(self.stats[0]['freq_max'])
        )
        msg = f'{freq:>14}'
        ret.append(self.curse_add_line(msg, 'DEFAULT'))

        # Third line is memory
        # Note: for the moment not available
        ret.append(self.curse_new_line())
        ret.append(self.curse_add_line('{:<12}'.format('mem:')))
        msg = self._format_value(self.stats[0]['mem'], unit='%')
        ret.append(
            self.curse_add_line(msg, self.get_views(key='mem', item=self.stats[0][self.get_key()], option='mem'))
        )

        # Fourth line is temperature
        # Note: for the moment only available for INTEL NPU
        ret.append(self.curse_new_line())
        ret.append(self.curse_add_line('{:<12}'.format('temperature:')))
        msg = self._format_value(self.stats[0]['temperature'], unit='C')
        ret.append(
            self.curse_add_line(
                msg, self.get_views(key='temperature', item=self.stats[0][self.get_key()], option='decoration')
            )
        )

        return ret



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Virsh (QEMU/KVM) Extension unit for Glances' Vms plugin."""

import os
import re
from functools import cache
from typing import Any

from glances.globals import nativestr
from glances.plugins.vms.engines import VmsExtension
from glances.secure import secure_popen

# Check if virsh binary exist
# TODO: make this path configurable from the Glances configuration file
VIRSH_PATH = '/usr/bin/virsh'
VIRSH_VERSION_OPTIONS = 'version'
VIRSH_INFO_OPTIONS = 'list --all'
VIRSH_DOMAIN_STATS_OPTIONS = 'domstats --nowait'
VIRSH_DOMAIN_TITLE_OPTIONS = 'desc --title'
import_virsh_error_tag = not os.path.exists(VIRSH_PATH) or not os.access(VIRSH_PATH, os.X_OK)


class VmExtension(VmsExtension):
    """Glances' Virsh Plugin's Vm Extension unit"""

    CONTAINER_ACTIVE_STATUS = ['running']

    def __init__(self):
        self.ext_name = "Virsh (Vm)"

    @cache
    def update_version(self):
        # > virsh version
        # Compiled against library: libvirt 10.0.0
        # Using library: libvirt 10.0.0
        # Using API: QEMU 10.0.0
        # Running hypervisor: QEMU 8.2.2
        ret_cmd = secure_popen(f'{VIRSH_PATH} {VIRSH_VERSION_OPTIONS}')
        try:
            ret = ret_cmd.splitlines()[3].split(':')[1].lstrip()
        except IndexError:
            return ''
        return ret

    def update_domains(self):
        # ❯ virsh list --all
        #  Id   Name              State
        # ----------------------------------
        #  4    win11             paused
        #  -    Kali_Linux_2024   shut off
        ret_cmd = secure_popen(f'{VIRSH_PATH} {VIRSH_INFO_OPTIONS}')

        try:
            # Ignore first two lines (header) and the last one (empty)
            lines = ret_cmd.splitlines()[2:-1]
        except IndexError:
            return {}

        ret = {}
        for line in lines:
            domain = re.split(r'\s{2,}', line)
            ret[domain[1]] = {
                'id': domain[0],
                'state': domain[2],
            }

        return ret

    def update_stats(self, domain):
        # ❯ virsh domstats win11
        # Domain: 'win11'
        #   state.state=1
        #   state.reason=1
        #   cpu.time=1702145606000
        #   cpu.user=1329954143000
        #   cpu.system=372191462000
        #   cpu.cache.monitor.count=0
        #   cpu.haltpoll.success.time=60243506159
        #   cpu.haltpoll.fail.time=34398762096
        #   balloon.current=8388608
        #   balloon.maximum=8388608
        #   balloon.last-update=0
        #   balloon.rss=8439116
        #   vcpu.current=4
        #   vcpu.maximum=4
        #   vcpu.0.state=1
        #   vcpu.0.time=955260000000
        #   vcpu.0.wait=0
        #   vcpu.0.delay=1305744538
        #   vcpu.0.halt_wakeup.sum=701415
        #   vcpu.0.halt_successful_poll.sum=457799
        #   vcpu.0.pf_mmio_spte_created.sum=38376
        #   vcpu.0.pf_emulate.sum=38873
        #   vcpu.0.fpu_reload.sum=8252258
        #   vcpu.0.insn_emulation.sum=6544119
        #   vcpu.0.signal_exits.sum=144
        #   vcpu.0.invlpg.sum=0
        #   vcpu.0.request_irq_exits.sum=0
        #   vcpu.0.preemption_reported.sum=25322
        #   vcpu.0.l1d_flush.sum=0
        #   vcpu.0.guest_mode.cur=no
        #   vcpu.0.halt_poll_fail_ns.sum=8369146975
        #   vcpu.0.pf_taken.sum=11455007
        #   vcpu.0.notify_window_exits.sum=0
        #   vcpu.0.directed_yield_successful.sum=0
        #   vcpu.0.host_state_reload.sum=8982225
        #   vcpu.0.nested_run.sum=0
        #   vcpu.0.nmi_injections.sum=1
        #   vcpu.0.pf_spurious.sum=7
        #   vcpu.0.halt_exits.sum=1173865
        #   vcpu.0.exits.sum=28094008
        #   vcpu.0.mmio_exits.sum=6536664
        #   vcpu.0.pf_fixed.sum=11410566
        #   vcpu.0.insn_emulation_fail.sum=0
        #   vcpu.0.io_exits.sum=1723037
        #   vcpu.0.halt_attempted_poll.sum=610576
        #   vcpu.0.req_event.sum=3245967
        #   vcpu.0.irq_exits.sum=1265655
        #   vcpu.0.blocking.cur=yes
        #   vcpu.0.irq_injections.sum=594
        #   vcpu.0.preemption_other.sum=11312
        #   vcpu.0.pf_fast.sum=5377009
        #   vcpu.0.hypercalls.sum=4799
        #   vcpu.0.nmi_window_exits.sum=0
        #   vcpu.0.directed_yield_attempted.sum=0
        #   vcpu.0.tlb_flush.sum=286197
        #   vcpu.0.halt_wait_ns.sum=872563463627
        #   vcpu.0.irq_window_exits.sum=0
        #   vcpu.0.pf_guest.sum=0
        #   vcpu.0.halt_poll_success_ns.sum=13527499051
        #   vcpu.0.halt_poll_invalid.sum=0
        #   vcpu.1...
        #   net.count=1
        #   net.0.name=vnet3
        #   net.0.rx.bytes=418454563
        #   net.0.rx.pkts=291468
        #   net.0.rx.errs=0
        #   net.0.rx.drop=7
        #   net.0.tx.bytes=3884562
        #   net.0.tx.pkts=35405
        #   net.0.tx.errs=0
        #   net.0.tx.drop=0
        #   block.count=2
        #   block.0.name=sda
        #   block.0.path=/var/lib/libvirt/images/win11.qcow2
        #   block.0.backingIndex=2
        #   block.0.rd.reqs=246802
        #   block.0.rd.bytes=18509028864
        #   block.0.rd.times=83676206416
        #   block.0.wr.reqs=471370
        #   block.0.wr.bytes=28260229120
        #   block.0.wr.times=158585666809
        #   block.0.fl.reqs=46269
        #   block.0.fl.times=181262854634
        #   block.0.allocation=137460187136
        #   block.0.capacity=137438953472
        #   block.0.physical=13457760256
        #   block.1...
        #   dirtyrate.calc_status=0
        #   dirtyrate.calc_start_time=0
        #   dirtyrate.calc_period=0
        #   dirtyrate.calc_mode=page-sampling
        #   vm.remote_tlb_flush.sum=436456
        #   vm.nx_lpage_splits.cur=0
        #   vm.pages_1g.cur=0
        #   vm.pages_2m.cur=859
        #   vm.pages_4k.cur=1196897
        #   vm.max_mmu_page_hash_collisions.max=0
        #   vm.mmu_pde_zapped.sum=0
        #   vm.max_mmu_rmap_size.max=0
        #   vm.mmu_cache_miss.sum=0
        #   vm.mmu_recycled.sum=0
        #   vm.mmu_unsync.cur=0
        #   vm.mmu_shadow_zapped.sum=0
        #   vm.mmu_flooded.sum=0
        #   vm.mmu_pte_write.sum=0
        #   vm.remote_tlb_flush_requests.sum=609455
        ret_cmd = secure_popen(f'{VIRSH_PATH} {VIRSH_DOMAIN_STATS_OPTIONS} {domain}')

        try:
            # Ignore first line (domain name already know) and last line (empty)
            lines = ret_cmd.splitlines()[1:-1]
        except IndexError:
            return {}

        ret = {}
        for line in lines:
            k, v = re.split(r'\s*=\s*', line.lstrip())
            ret[k] = v

        return ret

    @cache
    def update_title(self, domain):
        # ❯ virsh desc --title Kali_Linux_2024
        # Kali Linux 2024
        ret_cmd = secure_popen(f'{VIRSH_PATH} {VIRSH_DOMAIN_TITLE_OPTIONS} {domain}')

        return ret_cmd.rstrip()

    def update(self, all_tag) -> tuple[dict, list[dict]]:
        """Update Vm stats using the input method."""
        # Can not run virsh on this system then...
        if import_virsh_error_tag:
            return '', []

        # Update VM stats
        version_stats = self.update_version()
        domains_stats = self.update_domains()
        returned_stats = []
        for k, v in domains_stats.items():
            # Only display when VM in on 'running' and 'paused' states
            # Why paused ? Because a paused VM consume memory
            if all_tag or self._want_display(v, 'state', ['running', 'paused']):
                returned_stats.append(self.generate_stats(k, v))

        return version_stats, returned_stats

    @property
    def key(self) -> str:
        """Return the key of the list."""
        return 'name'

    def _want_display(self, domain_stats, key, values):
        return domain_stats.get(key).lower() in [v.lower() for v in values]

    def generate_stats(self, domain_name, domain_stats) -> dict[str, Any]:
        # Update stats and title for the domain
        vm_stats = self.update_stats(domain_name)
        vm_title = self.update_title(domain_name)

        return {
            'key': self.key,
            'name': nativestr(domain_name),
            'id': domain_stats.get('id'),
            'status': domain_stats.get('state').lower() if domain_stats.get('state') else None,
            'release': nativestr(vm_title),
            'cpu_count': int(vm_stats.get('vcpu.current', 1)) if vm_stats.get('vcpu.current', 1) else None,
            'cpu_time': int(vm_stats.get('cpu.time', 0)) / 1000000000 * 100,
            'memory_usage': int(vm_stats.get('balloon.rss')) * 1024 if vm_stats.get('balloon.rss') else None,
            'memory_total': int(vm_stats.get('balloon.maximum')) * 1024 if vm_stats.get('balloon.maximum') else None,
            'load_1min': None,
            'load_5min': None,
            'load_15min': None,
            'ipv4': None,
            # TODO: disk
        }



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""OpenTSDB interface class."""

import sys
from numbers import Number

import potsdb

from glances.exports.export import GlancesExport
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the OpenTSDB export module."""

    def __init__(self, config=None, args=None):
        """Init the OpenTSDB export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        # N/A

        # Optionals configuration keys
        self.prefix = None
        self.tags = None

        # Load the configuration file
        self.export_enable = self.load_conf('opentsdb', mandatories=['host', 'port'], options=['prefix', 'tags'])
        if not self.export_enable:
            exit('Missing OPENTSDB config')

        # Default prefix for stats is 'glances'
        if self.prefix is None:
            self.prefix = 'glances'

        # Init the OpenTSDB client
        self.client = self.init()

    def init(self):
        """Init the connection to the OpenTSDB server."""
        if not self.export_enable:
            return None

        try:
            db = potsdb.Client(self.host, port=int(self.port), check_host=True)
        except Exception as e:
            logger.critical(f"Cannot connect to OpenTSDB server {self.host}:{self.port} ({e})")
            sys.exit(2)

        return db

    def export(self, name, columns, points):
        """Export the stats to the OpenTSDB server."""
        for i in range(len(columns)):
            if not isinstance(points[i], Number):
                continue
            stat_name = f'{self.prefix}.{name}.{columns[i]}'
            stat_value = points[i]
            tags = self.parse_tags(self.tags)
            try:
                self.client.send(stat_name, stat_value, **tags)
            except Exception as e:
                logger.error(f"Can not export stats {name} to OpenTSDB ({e})")
        logger.debug(f"Export {name} stats to OpenTSDB")

    def exit(self):
        """Close the OpenTSDB export module."""
        # Waits for all outstanding metrics to be sent and background thread closes
        self.client.wait()
        # Call the father method
        super().exit()



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""MongoDB interface class."""

import sys
from urllib.parse import quote_plus

import pymongo

from glances.exports.export import GlancesExport
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the MongoDB export module."""

    def __init__(self, config=None, args=None):
        """Init the MongoDB export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        self.db = None

        # Optional configuration keys
        self.user = None
        self.password = None

        # Load the Cassandra configuration file section
        self.export_enable = self.load_conf('mongodb', mandatories=['host', 'port', 'db'], options=['user', 'password'])
        if not self.export_enable:
            sys.exit(2)

        # Init the CouchDB client
        self.client = self.init()

    def init(self):
        """Init the connection to the CouchDB server."""
        if not self.export_enable:
            return None

        server_uri = f'mongodb://{quote_plus(self.user)}:{quote_plus(self.password)}@{self.host}:{self.port}'

        try:
            client = pymongo.MongoClient(server_uri)
            client.admin.command('ping')
        except Exception as e:
            logger.critical(f"Cannot connect to MongoDB server {self.host}:{self.port} ({e})")
            sys.exit(2)
        else:
            logger.info("Connected to the MongoDB server")

        return client

    def database(self):
        """Return the CouchDB database object"""
        return self.client[self.db]

    def export(self, name, columns, points):
        """Write the points to the MongoDB server."""
        logger.debug(f"Export {name} stats to MongoDB")

        # Create DB input
        data = dict(zip(columns, points))

        # Write data to the MongoDB database
        try:
            self.database()[name].insert_one(data)
        except Exception as e:
            logger.error(f"Cannot export {name} stats to MongoDB ({e})")



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Manage the AMPs list."""

import os
import re
import threading

from glances.globals import amps_path, listkeys
from glances.logger import logger
from glances.processes import glances_processes


class AmpsList:
    """This class describes the optional application monitoring process list.

    The AMP list is a list of processes with a specific monitoring action.

    The list (Python list) is composed of items (Python dict).
    An item is defined (dict keys):
    *...
    """

    # The dict
    __amps_dict = {}

    def __init__(self, args, config):
        """Init the AMPs list."""
        self.args = args
        self.config = config

        # Load the AMP configurations / scripts
        self.load_configs()

    def load_configs(self):
        """Load the AMP configuration files."""
        if self.config is None:
            return False

        # For each AMP script, call the load_config method
        for s in self.config.sections():
            if s.startswith("amp_"):
                # An AMP section exists in the configuration file
                # If an AMP module exist in amps_path (glances/amps) folder then use it
                amp_name = s[4:]
                amp_module = os.path.join(amps_path, amp_name)
                if not os.path.exists(amp_module):
                    # If not, use the default script
                    amp_module = os.path.join(amps_path, "default")
                try:
                    amp = __import__(os.path.basename(amp_module))
                except ImportError as e:
                    logger.warning(f"Missing Python Lib ({e}), cannot load AMP {amp_name}")
                except Exception as e:
                    logger.warning(f"Cannot load AMP {amp_name} ({e})")
                else:
                    # Add the AMP to the dictionary
                    # The key is the AMP name
                    # for example, the file glances_xxx.py
                    # generate self._amps_list["xxx"] = ...
                    self.__amps_dict[amp_name] = amp.Amp(name=amp_name, args=self.args)
                    # Load the AMP configuration
                    self.__amps_dict[amp_name].load_config(self.config)
        # Log AMPs list
        logger.debug(f"AMPs list: {self.getList()}")

        return True

    def __str__(self):
        return str(self.__amps_dict)

    def __repr__(self):
        return self.__amps_dict

    def __getitem__(self, item):
        return self.__amps_dict[item]

    def __len__(self):
        return len(self.__amps_dict)

    def update(self):
        """Update the command result attributed."""
        # Get the current processes list (once)
        processlist = glances_processes.get_list()

        # Iter upon the AMPs dict
        for k, v in self.get().items():
            if not v.enable():
                # Do not update if the enable tag is set
                continue

            if v.regex() is None:
                # If there is no regex, execute anyway (see issue #1690)
                v.set_count(0)
                # Call the AMP update method
                thread = threading.Thread(target=v.update_wrapper, args=[[]])
                thread.start()
                continue

            amps_list = self._build_amps_list(v, processlist)

            if amps_list:
                # At least one process is matching the regex
                logger.debug(f"AMPS: {len(amps_list)} processes {k} detected ({amps_list})")
                # Call the AMP update method
                thread = threading.Thread(target=v.update_wrapper, args=[amps_list])
                thread.start()
            else:
                # Set the process number to 0
                v.set_count(0)
                if v.count_min() is not None and v.count_min() > 0:
                    # Only display the "No running process message" if count_min is defined
                    v.set_result("No running process")

        return self.__amps_dict

    def _build_amps_list(self, amp_value, processlist):
        """Return the AMPS process list according to the amp_value

        Search application monitored processes by a regular expression
        """
        try:
            # Search in both cmdline and name (for kernel thread, see #1261)
            ret = [
                {'pid': p['pid'], 'cpu_percent': p['cpu_percent'], 'memory_percent': p['memory_percent']}
                for p in processlist
                if re.search(amp_value.regex(), p['name'])
                or ((cmdline := p.get('cmdline')) and re.search(amp_value.regex(), ' '.join(cmdline)))
            ]

        except (TypeError, KeyError) as e:
            logger.debug(f"Can not build AMPS list ({e})")

        return ret

    def getList(self):
        """Return the AMPs list."""
        return listkeys(self.__amps_dict)

    def get(self):
        """Return the AMPs dict."""
        return self.__amps_dict

    def set(self, new_dict):
        """Set the AMPs dict."""
        self.__amps_dict = new_dict



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Fetch mode interface class."""

import jinja2

from glances import api
from glances.logger import logger

DEFAULT_FETCH_TEMPLATE = """

━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
✨ {{ gl.system['hostname'] }}{{ ' | ' + gl.ip['address'] if gl.ip['address'] else '' }} | Uptime: {{ gl.uptime }}
⚙️  {{ gl.system['hr_name'] }}

💡 LOAD     {{ '%0.2f'| format(gl.load['min1']) }}/min1 |\
 {{ '%0.2f'| format(gl.load['min5']) }}/min5 |\
 {{ '%0.2f'| format(gl.load['min15']) }}/min15
⚡ CPU      {{ gl.bar(gl.cpu['total']) }} {{ gl.cpu['total'] }}% of {{ gl.core['log'] }} cores
🧠 MEM      {{ gl.bar(gl.mem['percent']) }} {{ gl.mem['percent'] }}% ({{ gl.auto_unit(gl.mem['used']) }} /\
 {{ gl.auto_unit(gl.mem['total']) }})
{% for fs in gl.fs.keys() %}\
💾 {% if loop.index == 1 %}DISK{% else %}    {% endif %}\
     {{ gl.bar(gl.fs[fs]['percent']) }} {{ gl.fs[fs]['percent'] }}% ({{ gl.auto_unit(gl.fs[fs]['used']) }} /\
 {{ gl.auto_unit(gl.fs[fs]['size']) }}) for {{ fs }}
{% endfor %}\
{% for net in gl.network.keys() %}\
📡 {% if loop.index == 1 %}NET{% else %}   {% endif %}\
      ↓ {{ gl.auto_unit(gl.network[net]['bytes_recv_rate_per_sec']) }}b/s\
 ↑ {{ gl.auto_unit(gl.network[net]['bytes_sent_rate_per_sec']) }}b/s for {{ net }}
{% endfor %}\

🔥 TOP PROCESS by CPU
{% for process in gl.top_process() %}\
{{ loop.index }}️⃣ {{ process['name'][:20] }}{{ ' ' * (20 - process['name'][:20] | length) }}\
    ⚡ {{ process['cpu_percent'] }}% CPU\
{{ ' ' * (8 - (gl.auto_unit(process['cpu_percent']) | length)) }}\
    🧠 {{ gl.auto_unit(process['memory_info']['rss']) }}B MEM
{% endfor %}\
🔥 TOP PROCESS by MEM
{% for process in gl.top_process(sorted_by='memory_percent', sorted_by_secondary='cpu_percent') %}\
{{ loop.index }}️⃣ {{ process['name'][:20] }}{{ ' ' * (20 - process['name'][:20] | length) }}\
    🧠 {{ gl.auto_unit(process['memory_info']['rss']) }}B MEM\
{{ ' ' * (7 - (gl.auto_unit(process['memory_info']['rss']) | length)) }}\
    ⚡ {{ process['cpu_percent'] }}% CPU
{% endfor %}\
━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━

"""


class GlancesStdoutFetch:
    """This class manages the Stdout JSON display."""

    def __init__(self, config=None, args=None):
        # Init
        self.config = config
        self.args = args
        self.gl = api.GlancesAPI(self.config, self.args)

    def end(self):
        pass

    def update(self, stats, duration=3, cs_status=None, return_to_browser=False):
        """Display fetch from the template file to stdout."""
        if self.args.fetch_template == "":
            fetch_template = DEFAULT_FETCH_TEMPLATE
        else:
            logger.info("Using fetch template file: " + self.args.fetch_template)
            # Load the template from the file given in the self.args.fetch_template argument
            with open(self.args.fetch_template) as f:
                fetch_template = f.read()

        # Create a Jinja2 environment
        jinja_env = jinja2.Environment(loader=jinja2.BaseLoader(), autoescape=True)
        template = jinja_env.from_string(fetch_template)
        output = template.render(gl=self.gl)
        print(output)

        # Return True to exit directly (no refresh)
        return True



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""
I am your father...

...of all Glances model plugins.
"""

import copy
import re
from datetime import datetime

from glances.actions import GlancesActions
from glances.events_list import glances_events
from glances.globals import (
    auto_unit,
    dictlist,
    dictlist_json_dumps,
    json_dumps,
    list_to_dict,
    listkeys,
    mean,
    nativestr,
    split_esc,
)
from glances.history import GlancesHistory
from glances.logger import logger
from glances.outputs.glances_unicode import unicode_message
from glances.thresholds import glances_thresholds
from glances.timer import Counter, Timer, getTimeSinceLastUpdate

fields_unit_short = {'percent': '%'}

fields_unit_type = {
    'percent': 'float',
    'percents': 'float',
    'number': 'int',
    'numbers': 'int',
    'int': 'int',
    'ints': 'int',
    'float': 'float',
    'floats': 'float',
    'second': 'int',
    'seconds': 'int',
    'byte': 'int',
    'bytes': 'int',
}


class GlancesPluginModel:
    """Main class for Glances plugin model."""

    def __init__(self, args=None, config=None, items_history_list=None, stats_init_value={}, fields_description=None):
        """Init the plugin of plugins model class.

        All Glances' plugins model should inherit from this class. Most of the
        methods are already implemented in the father classes.

        Your plugin should return a dict or a list of dicts (stored in the
        self.stats). As an example, you can have a look on the mem plugin
        (for dict) or network (for list of dicts).

        From version 4 of the API, the plugin should return a dict.

        A plugin should implement:
        - the reset method: to set your self.stats variable to {} or []
        - the update method: where your self.stats variable is set
        and optionally:
        - the get_key method: set the key of the dict (only for list of dict)
        - all others methods you want to overwrite

        :args: args parameters
        :config: configuration parameters
        :items_history_list: list of items to store in the history
        :stats_init_value: Default value for a stats item
        """
        # Build the plugin name
        # Internal or external module (former prefixed by 'glances.plugins')
        _mod = self.__class__.__module__.replace('glances.plugins.', '')
        self.plugin_name = _mod.split('.')[0]

        if self.plugin_name.startswith('glances_'):
            self.plugin_name = self.plugin_name.split('glances_')[1]
        logger.debug(f"Init {self.plugin_name} plugin")

        # Init the args
        self.args = args

        # Init the default alignment (for curses)
        self._align = 'left'

        # Init the input method
        self._input_method = 'local'
        self._short_system_name = None

        # Init the history list
        self.items_history_list = items_history_list
        self.stats_history = self.init_stats_history()

        # Init the limits (configuration keys) dictionary
        self._limits = {}
        if config is not None:
            logger.debug(f'Load section {self.plugin_name} in Glances configuration file')
            self.load_limits(config=config)

        # Init the alias (dictionary)
        self.alias = self.read_alias()

        # Init the actions
        self.actions = GlancesActions(args=args)

        # Init the views
        self.views = {}

        # Hide stats if all the hide_zero_fields has never been != 0
        # Default is False, always display stats
        self.hide_zero = False
        # The threshold needed to display a value if hide_zero is true.
        # Only hide a value if it is less than hide_threshold_bytes.
        self.hide_threshold_bytes = 0
        self.hide_zero_fields = []

        # Set the initial refresh time to display stats the first time
        self.refresh_timer = Timer(0)

        # Init stats description
        self.fields_description = fields_description

        # Init MMM (Min/Max/Mean) tracking for fields with mmm=True
        self._mmm_fields = self._init_mmm_fields()

        # Init the stats
        self.stats_init_value = stats_init_value
        self.time_since_last_update = None
        self.stats = None
        self.stats_previous = None
        self.reset()

    def __str__(self):
        """Return the human-readable stats."""
        return str(self.stats)

    def __repr__(self):
        """Return the raw stats."""
        if isinstance(self.stats, list):
            return str(list_to_dict(self.stats))
        return str(self.stats)

    def __getitem__(self, item):
        """Return the stats item."""
        if isinstance(self.stats, dict) and item in self.stats:
            return self.stats[item]

        if isinstance(self.stats, list):
            ltd = list_to_dict(self.stats)
            if item in ltd:
                return ltd[item]

        raise KeyError(f"'{self.__class__.__name__}' object has no key '{item}'")

    def keys(self):
        """Return the keys of the stats."""
        if isinstance(self.stats, dict):
            return listkeys(self.stats)
        if isinstance(self.stats, list):
            return listkeys(list_to_dict(self.stats))
        return []

    def get(self, item, default=None):
        """Return the stats item or default if not found."""
        try:
            return self[item]
        except KeyError:
            return default

    def get_init_value(self):
        """Return a copy of the init value."""
        return copy.copy(self.stats_init_value)

    def _init_mmm_fields(self):
        """Initialize MMM (Min/Max/Mean) field tracking.

        Scan fields_description for fields with mmm=True and create tracking structures.
        Also automatically add field descriptions for the generated fields.

        Returns a dictionary with mmm field tracking info.
        """
        mmm_fields = {}

        if self.fields_description is None:
            return mmm_fields

        # Find all fields with mmm=True (iterate over keys to avoid dict size change during iteration)
        mmm_field_names = [
            field_name for field_name, field_info in self.fields_description.items() if field_info.get('mmm', False)
        ]

        # Now process the mmm fields
        for field_name in mmm_field_names:
            field_info = self.fields_description[field_name]

            # Initialize tracking for this field
            mmm_fields[field_name] = {
                'values': [],  # Keep history for mean calculation
                'min': None,
                'max': None,
                'unit': field_info.get('unit', ''),
            }

            # Automatically add field descriptions for the generated fields if not already present
            suffix_map = {
                '_min': f"Minimum {field_name} observed since Glances startup.",
                '_max': f"Maximum {field_name} observed since Glances startup.",
                '_mean': f"Mean (average) {field_name} computed from the history.",
            }

            for suffix, description in suffix_map.items():
                generated_field = field_name + suffix
                if generated_field not in self.fields_description:
                    self.fields_description[generated_field] = {
                        'description': description,
                        'unit': field_info.get('unit', ''),
                    }

        return mmm_fields

    def _update_mmm_fields(self, stats):
        """Update MMM (Min/Max/Mean) fields for all fields with mmm=True.

        This method should be called after the plugin's update method.
        It will compute and add _min, _max, and _mean fields to stats.

        Args:
            stats: The stats dictionary to update (should be a dict or dict from list item)

        Returns:
            The stats dictionary with mmm fields added
        """
        if not isinstance(stats, dict) or not self._mmm_fields:
            return stats

        # Update mmm fields for each tracked field
        for field_name, mmm_info in self._mmm_fields.items():
            if field_name not in stats:
                continue

            current_value = stats[field_name]

            # Only process numeric values
            if current_value is None or (not isinstance(current_value, (int, float))):
                continue

            # Keep history for mean calculation (limit to reasonable size to avoid memory growth)
            max_history_size = 28800  # ~1 day at 1 sample/sec
            mmm_info['values'].append(current_value)
            if len(mmm_info['values']) > max_history_size:
                mmm_info['values'].pop(0)

            # Update min and max
            if mmm_info['min'] is None or current_value < mmm_info['min']:
                mmm_info['min'] = current_value
            if mmm_info['max'] is None or current_value > mmm_info['max']:
                mmm_info['max'] = current_value

            # Add generated fields to stats
            stats[field_name + '_min'] = mmm_info['min']
            stats[field_name + '_max'] = mmm_info['max']

            # Compute mean from history
            if mmm_info['values']:
                stats[field_name + '_mean'] = round(mean(mmm_info['values']), 2)

        return stats

    def _update_mmm_fields_on_list(self, stats_list):
        """Update MMM fields for a list of stats dictionaries.

        Args:
            stats_list: A list of stats dictionaries

        Returns:
            The list with mmm fields updated for each item
        """
        if not isinstance(stats_list, list):
            return stats_list

        for stat in stats_list:
            if isinstance(stat, dict):
                self._update_mmm_fields(stat)

        return stats_list

    def reset(self):
        """Reset the stats.

        This method should be overwritten by child classes.
        """
        self.stats = self.get_init_value()

    def exit(self):
        """Just log an event when Glances exit."""
        logger.debug(f"Stop the {self.plugin_name} plugin")

    def get_key(self):
        """Return the key of the list."""
        return

    def is_enabled(self, plugin_name=None):
        """Return true if plugin is enabled."""
        if not plugin_name:
            plugin_name = self.plugin_name
        try:
            d = getattr(self.args, 'disable_' + plugin_name)
        except AttributeError:
            d = getattr(self.args, 'enable_' + plugin_name, True)
        return d is False

    def is_disabled(self, plugin_name=None):
        """Return true if plugin is disabled."""
        return not self.is_enabled(plugin_name=plugin_name)

    def history_enable(self):
        return self.args is not None and not self.args.disable_history and self.get_items_history_list() is not None

    def init_stats_history(self):
        """Init the stats history (dict of GlancesAttribute)."""
        if self.history_enable():
            init_list = [a['name'] for a in self.get_items_history_list()]
            logger.debug(f"Stats history activated for plugin {self.plugin_name} (items: {init_list})")
        return GlancesHistory()

    def reset_stats_history(self):
        """Reset the stats history (dict of GlancesAttribute)."""
        if self.history_enable():
            reset_list = [a['name'] for a in self.get_items_history_list()]
            logger.debug(f"Reset history for plugin {self.plugin_name} (items: {reset_list})")
            self.stats_history.reset()

    def update_stats_history(self):
        """Update stats history."""
        # Exit if no history
        if not self.history_enable():
            return
        # Build the history
        _get_export = self.get_export()
        if not (_get_export and self.history_enable()):
            return
        # Itern through items history
        item_name = '' if self.get_key() is None else self.get_key()
        for i in self.get_items_history_list():
            if isinstance(_get_export, list):
                # Stats is a list of data
                # Iter through stats (for example, iter through network interface)
                for l_export in _get_export:
                    if i['name'] in l_export:
                        self.stats_history.add(
                            nativestr(l_export[item_name]) + '_' + nativestr(i['name']),
                            l_export[i['name']],
                            description=i['description'],
                            history_max_size=self._limits['history_size'],
                        )
            else:
                # Stats is not a list
                # Add the item to the history directly
                self.stats_history.add(
                    nativestr(i['name']),
                    _get_export[i['name']],
                    description=i['description'],
                    history_max_size=self._limits['history_size'],
                )

    def get_items_history_list(self):
        """Return the items history list."""
        return self.items_history_list

    def get_raw_history(self, item=None, nb=0):
        """Return the history (RAW format).

        - the stats history (dict of list) if item is None
        - the stats history for the given item (list) instead
        - None if item did not exist in the history
        """
        s = self.stats_history.get(nb=nb)
        if item is None:
            return s
        if item in s:
            return s[item]
        return None

    def get_export_history(self, item=None):
        """Return the stats history object to export."""
        return self.get_raw_history(item=item)

    def get_stats_history(self, item=None, nb=0):
        """Return the stats history (JSON format)."""
        s = self.stats_history.get_json(nb=nb)

        if item is None:
            return json_dumps(s)

        return dictlist_json_dumps(s, item)

    def get_trend(self, item, nb=30):
        """Get the trend regarding to the last nb values.

        The trend is the diffirence between the mean of the last 0 to nb / 2
        and nb / 2 to nb values.
        """
        raw_history = self.get_raw_history(item=item, nb=nb)
        if raw_history is None or len(raw_history) < nb:
            return None
        last_nb = [v[1] for v in raw_history]
        return mean(last_nb[nb // 2 :]) - mean(last_nb[: nb // 2])

    @property
    def input_method(self):
        """Get the input method."""
        return self._input_method

    @input_method.setter
    def input_method(self, input_method):
        """Set the input method.

        * local: system local grab (psutil or direct access)
        * snmp: Client server mode via SNMP
        * glances: Client server mode via Glances API
        """
        self._input_method = input_method

    @property
    def short_system_name(self):
        """Get the short detected OS name (SNMP)."""
        return self._short_system_name

    def sorted_stats(self):
        """Get the stats sorted by an alias (if present) or key."""
        key = self.get_key()
        if key is None:
            return self.stats
        try:
            return sorted(
                self.stats,
                key=lambda stat: tuple(
                    int(part) if part.isdigit() else part.lower()
                    for part in re.split(r"(\d+|\D+)", self.has_alias(stat[key]) or stat[key])
                ),
            )
        except TypeError:
            # Correct "Starting an alias with a number causes a crash #1885"
            return sorted(
                self.stats,
                key=lambda stat: tuple(
                    part.lower() for part in re.split(r"(\d+|\D+)", self.has_alias(stat[key]) or stat[key])
                ),
            )

    @short_system_name.setter
    def short_system_name(self, short_name):
        """Set the short detected OS name (SNMP)."""
        self._short_system_name = short_name

    def set_stats(self, input_stats):
        """Set the stats to input_stats."""
        self.stats = input_stats

    def get_stats_snmp(self, bulk=False, snmp_oid=None):
        """Update stats using SNMP.

        If bulk=True, use a bulk request instead of a get request.
        """
        snmp_oid = snmp_oid or {}

        from glances.snmp import GlancesSNMPClient

        # Init the SNMP request
        snmp_client = GlancesSNMPClient(
            host=self.args.client,
            port=self.args.snmp_port,
            version=self.args.snmp_version,
            community=self.args.snmp_community,
        )

        # Process the SNMP request
        ret = {}
        if bulk:
            # Bulk request
            snmp_result = snmp_client.getbulk_by_oid(0, 10, *list(snmp_oid.values()))
            logger.info(snmp_result)
            if len(snmp_oid) == 1:
                # Bulk command for only one OID
                # Note: key is the item indexed but the OID result
                for item in snmp_result:
                    if item.keys()[0].startswith(snmp_oid.values()[0]):
                        ret[snmp_oid.keys()[0] + item.keys()[0].split(snmp_oid.values()[0])[1]] = item.values()[0]
            else:
                # Build the internal dict with the SNMP result
                # Note: key is the first item in the snmp_oid
                index = 1
                for item in snmp_result:
                    item_stats = {}
                    item_key = None
                    for key in snmp_oid:
                        oid = snmp_oid[key] + '.' + str(index)
                        if oid in item:
                            if item_key is None:
                                item_key = item[oid]
                            else:
                                item_stats[key] = item[oid]
                    if item_stats:
                        ret[item_key] = item_stats
                    index += 1
        else:
            # Simple get request
            snmp_result = snmp_client.get_by_oid(*list(snmp_oid.values()))

            # Build the internal dict with the SNMP result
            for key in snmp_oid:
                ret[key] = snmp_result[snmp_oid[key]]

        return ret

    def get_raw(self):
        """Return the stats object."""
        return self.stats

    def get_api(self):
        """Return the stats object for the API.
        By default, return the raw stats."""
        return self.get_raw()

    def get_export(self):
        """Return the stats object to export.
        By default, return the raw stats.
        Note: this method could be overwritten by the plugin if a specific format is needed (ex: processlist)
        """
        return self.get_raw()

    def get_stats(self):
        """Return the stats object in JSON format."""
        return json_dumps(self.get_raw())

    def get_json(self):
        """Return the stats object in JSON format."""
        return self.get_stats()

    def get_raw_stats_item(self, item):
        """Return the stats object for a specific item in RAW format.

        Stats should be a list of dict (processlist, network...)
        """
        return dictlist(self.get_raw(), item)

    def get_raw_stats_key(self, item, key):
        """Return the stats object for a specific item in RAW format.

        Stats should be a list of dict (processlist, network...)
        """
        return {item: [i for i in self.get_raw() if 'key' in i and i[i['key']] == key][0].get(item)}

    def get_stats_item(self, item):
        """Return the stats object for a specific item in JSON format.

        Stats should be a list of dict (processlist, network...)
        """
        return dictlist_json_dumps(self.get_raw(), item)

    def get_raw_stats_value(self, item, value):
        """Return the stats object for a specific item=value.

        Return None if the item=value does not exist
        Return None if the item is not a list of dict
        """
        if not isinstance(self.get_raw(), list):
            return None

        if (not isinstance(value, int) and not isinstance(value, float)) and value.isdigit():
            value = int(value)
        try:
            return {value: [i for i in self.get_raw() if i[item] == value]}
        except (KeyError, ValueError) as e:
            logger.error(f"Cannot get item({item})=value({value}) ({e})")
            return None

    def get_stats_value(self, item, value):
        """Return the stats object for a specific item=value in JSON format.

        Stats should be a list of dict (processlist, network...)
        """
        rsv = self.get_raw_stats_value(item, value)
        if rsv is None:
            return None
        return json_dumps(rsv)

    def get_item_info(self, item, key, default=None):
        """Return the item info grabbed into self.fields_description."""
        if self.fields_description is None or item not in self.fields_description:
            return default
        return self.fields_description[item].get(key, default)

    def _build_field_decoration(self, field):
        """Return the field decoration.

        The decoration is used to display the field in the UI.
        """
        # Manage the decoration
        if self.fields_description and field in self.fields_description:
            if (
                self.fields_description[field].get('rate') is True
                and isinstance(self.stats, dict)
                and self.stats.get('time_since_update', 0) == 0
            ):
                return 'DEFAULT'
            if self.fields_description[field].get('log') is True:
                return self.get_alert_log(self.stats[field], header=field)
            if self.fields_description[field].get('alert') is True:
                return self.get_alert(self.stats[field], header=field)
        return 'DEFAULT'

    def _build_field_optional(self, field):
        """Return true if the field is optional."""
        if self.fields_description and field in self.fields_description:
            return self.fields_description[field].get('optional', False)
        return False

    def _build_view_for_field(self, key=None, field=None):
        view = {
            'decoration': self._build_field_decoration(field),
            'optional': self._build_field_optional(field),
            'additional': False,
            'splittable': False,
            'hidden': False,
        }

        # Manage the hidden feature
        # Allow to automatically hide fields when values is never different than 0
        # Refactoring done for #2929
        if not self.hide_zero:
            view['hidden'] = False
        elif key and key in self.views and field in self.views[key] and 'hidden' in self.views[key][field]:
            view['hidden'] = self.views[key][field]['hidden']
            if (
                field in self.hide_zero_fields
                and self.get_raw_stats_key(item=field, key=key).get(field) >= self.hide_threshold_bytes
            ):
                view['hidden'] = False
            # logger.info(f'{key=} {field=} {view["hidden"]=}')
        else:
            view['hidden'] = field in self.hide_zero_fields

        return view

    def update_views(self):
        """Update the stats views.

        The V of MVC
        A dict of dict with the needed information to display the stats.
        Example for the stat xxx:
        'xxx': {'decoration': 'DEFAULT',  >>> The decoration of the stats
                'optional': False,        >>> Is the stat optional
                'additional': False,      >>> Is the stat provide additional information
                'splittable': False,      >>> Is the stat can be cut (like process lon name)
                'hidden': False}          >>> Is the stats should be hidden in the UI
        """
        ret = {}

        if self.get_raw() is not None and isinstance(self.get_raw(), list) and self.get_key() is not None:
            # Stats are stored in a list of dict (ex: DISKIO, NETWORK, FS...)
            for i in self.get_raw():
                key = i[self.get_key()]
                ret[key] = {}
                for field in listkeys(i):
                    ret[key][field] = self._build_view_for_field(key=key, field=field)
        elif isinstance(self.get_raw(), dict) and self.get_raw() is not None:
            # Stats are stored in a dict (ex: CPU, LOAD...)
            for field in listkeys(self.get_raw()):
                ret[field] = self._build_view_for_field(key=None, field=field)

        self.views = ret

        return self.views

    def set_views(self, input_views):
        """Set the views to input_views."""
        self.views = input_views

    def reset_views(self):
        """Reset the views to input_views."""
        self.views = {}

    def get_views(self, item=None, key=None, option=None):
        """Return the views object.

        If key is None, return all the view for the current plugin
        else if option is None return the view for the specific key (all option)
        else return the view of the specific key/option

        Specify item if the stats are stored in a dict of dict (ex: NETWORK, FS...)
        """
        if item is None:
            item_views = self.views
        else:
            item_views = self.views[item]
        if key is None:
            return item_views
        if key not in item_views:
            return 'DEFAULT'
        if option is None:
            return item_views[key]
        if option in item_views[key]:
            return item_views[key][option]
        return 'DEFAULT'

    def get_json_views(self, item=None, key=None, option=None):
        """Return the views (in JSON)."""
        return json_dumps(self.get_views(item, key, option))

    def load_limits(self, config):
        """Load limits from the configuration file, if it exists."""
        # By default set the history length to 3 points per second during one day
        self._limits['history_size'] = 28800

        if not hasattr(config, 'has_section'):
            return False

        # Read the global section
        # TODO: not optimized because this section is loaded for each plugin...
        if config.has_section('global'):
            self._limits['history_size'] = config.get_float_value('global', 'history_size', default=28800)
            logger.debug("Load configuration key: {} = {}".format('history_size', self._limits['history_size']))

        # Read the plugin specific section
        if config.has_section(self.plugin_name):
            for level, _ in config.items(self.plugin_name):
                # Read limits
                limit = '_'.join([self.plugin_name, level])
                try:
                    self._limits[limit] = config.get_float_value(self.plugin_name, level)
                except ValueError:
                    self._limits[limit] = config.get_value(self.plugin_name, level).split(",")
                logger.debug(f"Load limit: {limit} = {self._limits[limit]}")

        return True

    @property
    def limits(self):
        """Return the limits object."""
        return self._limits

    @limits.setter
    def limits(self, input_limits):
        """Set the limits to input_limits."""
        self._limits = input_limits

    def set_refresh(self, value):
        """Set the plugin refresh rate"""
        self.set_limits('refresh', value)

    def get_refresh(self):
        """Return the plugin refresh time"""
        ret = self.get_limits(item='refresh')
        if ret is None:
            ret = self.args.time if hasattr(self.args, 'time') else 2
        return ret

    def get_refresh_time(self):
        """Return the plugin refresh time"""
        return self.get_refresh()

    def set_limits(self, item, value):
        """Set the limits object."""
        self._limits[f'{self.plugin_name}_{item}'] = value

    def get_limits(self, item=None):
        """Return the limits object."""
        if item is None:
            return self._limits
        return self._limits.get(f'{self.plugin_name}_{item}', None)

    def get_stats_action(self):
        """Return stats for the action.

        By default return all the stats.
        Can be overwrite by plugins implementation.
        For example, Docker will return self.stats['containers']
        """
        return self.stats

    def get_stat_name(self, header=None, action_key=None):
        """Return the stat name with an optional action_key and header"""
        ret = self.plugin_name
        if action_key is not None and action_key != '':
            ret += '_' + action_key
        if header is not None and header != '':
            ret += '_' + header
        return ret

    def get_alert(
        self,
        current=0,
        minimum=0,
        maximum=100,
        highlight_zero=True,
        is_max=False,
        header=None,
        action_key=None,
        log=False,
    ):
        """Return the alert status relative to a current value.

        Use this function for minor stats.

        If current < CAREFUL of max then alert = OK
        If current > CAREFUL of max then alert = CAREFUL
        If current > WARNING of max then alert = WARNING
        If current > CRITICAL of max then alert = CRITICAL

        If highlight=True than 0.0 is highlighted

        If defined 'header' is added between the plugin name and the status.
        Only useful for stats with several alert status.

        If defined, 'action_key' define the key for the actions.
        By default, the action_key is equal to the header.

        If log=True than add log if necessary
        elif log=False than do not log
        elif log=None than apply the config given in the conf file
        """
        # Manage 0 (0.0) value if highlight_zero is not True
        if not highlight_zero and current == 0:
            return 'DEFAULT'

        # Compute the %
        try:
            value = (current * 100) / maximum
        except ZeroDivisionError:
            return 'DEFAULT'
        except TypeError:
            return 'DEFAULT'

        # Build the stat_name
        stat_name = self.get_stat_name(header=header, action_key=action_key).lower()

        # Manage limits
        # If is_max is set then default style is set to MAX else default is set to OK
        ret = 'MAX' if is_max else 'OK'

        # Iter through limits
        critical = self.get_limit('critical', stat_name=stat_name)
        warning = self.get_limit('warning', stat_name=stat_name)
        careful = self.get_limit('careful', stat_name=stat_name)
        if critical and value >= critical:
            ret = 'CRITICAL'
        elif warning and value >= warning:
            ret = 'WARNING'
        elif careful and value >= careful:
            ret = 'CAREFUL'
        elif not careful and not warning and not critical:
            ret = 'DEFAULT'
        else:
            ret = 'OK'

        if current < minimum:
            ret = 'CAREFUL'

        # Manage log
        log_str = ""
        if self.get_limit_log(stat_name=stat_name, default_action=log) and ret != 'DEFAULT':
            # Add _LOG to the return string
            # So stats will be highlighted with a specific color
            log_str = "_LOG"
            # Add the log to the events list
            glances_events.add(ret, stat_name.upper(), value)

        # Manage threshold
        self.manage_threshold(stat_name, ret)

        # Manage action
        self.manage_action(stat_name, ret.lower(), header, action_key)

        # Default is 'OK'
        return ret + log_str

    def filter_stats(self, stats):
        """Filter the stats to keep only the fields we want (the one defined in fields_description)."""
        if hasattr(stats, '_asdict'):
            return {k: v for k, v in stats._asdict().items() if k in self.fields_description}
        if isinstance(stats, dict):
            return {k: v for k, v in stats.items() if k in self.fields_description}
        if isinstance(stats, list):
            return [self.filter_stats(s) for s in stats]
        return stats

    def manage_threshold(self, stat_name, trigger):
        """Manage the threshold for the current stat."""
        glances_thresholds.add(stat_name, trigger)

    def manage_action(self, stat_name, trigger, header, action_key):
        """Manage the action for the current stat."""
        # Here is a command line for the current trigger ?
        command, repeat = self.get_limit_action(trigger, stat_name=stat_name)
        if not command and not repeat:
            # Reset the trigger
            self.actions.set(stat_name, trigger)
        else:
            # Define the action key for the stats dict
            # If not define, then it sets to header
            if action_key is None:
                action_key = header

            # A command line is available for the current alert
            # 1) Build the {{mustache}} dictionary
            stats_action = copy.deepcopy(self.get_stats_action())
            if isinstance(stats_action, list):
                # If the stats are stored in a list of dict (fs plugin for example)
                mustache_dict = {}
                for item in stats_action:
                    # Add the limit to the mustache dict
                    item['critical'] = self.get_limit('critical', stat_name=stat_name)
                    item['warning'] = self.get_limit('warning', stat_name=stat_name)
                    item['careful'] = self.get_limit('careful', stat_name=stat_name)
                    # Add the current time (now)
                    item['time'] = datetime.now().isoformat()
                    if item[self.get_key()] == action_key:
                        mustache_dict = item
                        break
            else:
                # Use the stats dict
                # Add the limit to the mustache dict
                stats_action['critical'] = self.get_limit('critical', stat_name=stat_name)
                stats_action['warning'] = self.get_limit('warning', stat_name=stat_name)
                stats_action['careful'] = self.get_limit('careful', stat_name=stat_name)
                # Add the current time (now)
                stats_action['time'] = datetime.now().isoformat()
                mustache_dict = stats_action
            # 2) Run the action
            self.actions.run(stat_name, trigger, command, repeat, mustache_dict=mustache_dict)

    def get_alert_log(self, current=0, minimum=0, maximum=100, header="", action_key=None):
        """Get the alert log."""
        return self.get_alert(
            current=current, minimum=minimum, maximum=maximum, header=header, action_key=action_key, log=True
        )

    def is_limit(self, criticality, stat_name=""):
        """Return true if the criticality limit exist for the given stat_name"""
        return self.get_stat_name(stat_name).lower() + '_' + criticality in self._limits

    def get_limit(self, criticality=None, stat_name=""):
        """Return the limit value for the given criticality.
        If criticality is None, return the dict of all the limits."""
        if criticality is None:
            return self._limits

        # Get the limit for stat + header
        # Example: network_wlan0_rx_careful
        stat_name = stat_name.lower()
        if stat_name + '_' + criticality in self._limits:
            return self._limits[stat_name + '_' + criticality]
        if self.plugin_name + '_' + criticality in self._limits:
            return self._limits[self.plugin_name + '_' + criticality]

        return None

    def get_limit_action(self, criticality, stat_name=""):
        """Return the tuple (action, repeat) for the alert.

        - action is a command line
        - repeat is a bool
        """
        # Get the action for stat + header
        # Example: network_wlan0_rx_careful_action
        # Action key available ?
        ret = [
            (stat_name + '_' + criticality + '_action', False),
            (stat_name + '_' + criticality + '_action_repeat', True),
            (self.plugin_name + '_' + criticality + '_action', False),
            (self.plugin_name + '_' + criticality + '_action_repeat', True),
        ]
        for r in ret:
            if r[0] in self._limits:
                return self._limits[r[0]], r[1]

        # No key found, return None
        return None, None

    def get_limit_log(self, stat_name, default_action=False):
        """Return the log tag for the alert."""
        # Get the log tag for stat + header
        # Example: network_wlan0_rx_log
        if stat_name + '_log' in self._limits:
            return self._limits[stat_name + '_log'][0].lower() == 'true'
        if self.plugin_name + '_log' in self._limits:
            return self._limits[self.plugin_name + '_log'][0].lower() == 'true'
        return default_action

    def get_conf_value(self, value, header="", plugin_name=None, convert_bool=False, default=[]):
        """Return the configuration (header_) value for the current plugin.

        ...or the one given by the plugin_name var.
        """
        if plugin_name is None:
            # If not default use the current plugin name
            plugin_name = self.plugin_name

        if header != "":
            # Add the header
            plugin_name = plugin_name + '_' + header

        try:
            ret = self._limits[plugin_name + '_' + value]
            return bool(ret[0]) if convert_bool else ret
        except KeyError:
            return default

    def is_show(self, value, header=""):
        """Return True if the value is in the show configuration list.

        If the show value is empty, return True (show by default)

        The show configuration list is defined in the glances.conf file.
        It is a comma-separated list of regexp.

        Example for diskio:
        show=sda.*
        """
        return any(
            re.fullmatch(i, value, re.I)
            or (self.has_alias(value) is not None and re.fullmatch(i, self.has_alias(value), re.I))
            for i in self.get_conf_value('show', header=header)
        )

    def is_hide(self, value, header=""):
        """Return True if the value is in the hide configuration list.

        The hide configuration list is defined in the glances.conf file.
        It is a comma-separated list of regexp.

        Example for diskio:
        hide=sda2,sda5,loop.*
        """
        return any(
            re.fullmatch(i, value, re.I)
            or (self.has_alias(value) is not None and re.fullmatch(i, self.has_alias(value), re.I))
            for i in self.get_conf_value('hide', header=header)
        )

    def is_display(self, value, header=""):
        """Return True if the value should be displayed in the UI"""
        if self.get_conf_value('show', header=header) != []:
            return self.is_show(value, header=header)
        return not self.is_hide(value, header=header)

    def is_display_any(self, *values, header=""):
        """Return True if any of the values should be displayed in the UI"""
        if self.get_conf_value('show', header=header) != []:
            return any(self.is_show(value, header=header) for value in values)
        return not any(self.is_hide(value, header=header) for value in values)

    def read_alias(self):
        if self.plugin_name + '_' + 'alias' in self._limits:
            return {
                split_esc(i, ':')[0].lower(): split_esc(i, ':')[1]
                for i in self._limits[self.plugin_name + '_' + 'alias']
            }
        return {}

    def has_alias(self, header):
        """Return the alias name for the relative header it it exists otherwise None."""
        return self.alias.get(header, None)

    def msg_curse(self, args=None, max_width=None):
        """Return default string to display in the curse interface."""
        return [self.curse_add_line(str(self.stats))]

    def get_stats_display(self, args=None, max_width=None):
        """Return a dict with all the information needed to display the stat.

        key     | description
        ----------------------------
        display | Display the stat (True or False)
        msgdict | Message to display (list of dict [{ 'msg': msg, 'decoration': decoration } ... ])
        align   | Message position
        """
        display_curse = False

        if hasattr(self, 'display_curse'):
            display_curse = self.display_curse
        if hasattr(self, 'align'):
            align_curse = self._align

        if max_width is not None:
            ret = {'display': display_curse, 'msgdict': self.msg_curse(args, max_width=max_width), 'align': align_curse}
        else:
            ret = {'display': display_curse, 'msgdict': self.msg_curse(args), 'align': align_curse}

        return ret

    def curse_add_line(self, msg, decoration="DEFAULT", optional=False, additional=False, splittable=False):
        """Return a dict with.

        Where:
            msg: string
            decoration:
                DEFAULT: no decoration
                UNDERLINE: underline
                BOLD: bold
                TITLE: for stat title
                PROCESS: for process name
                STATUS: for process status
                CPU_TIME: for process cpu time
                OK: Value is OK and non logged
                OK_LOG: Value is OK and logged
                CAREFUL: Value is CAREFUL and non logged
                CAREFUL_LOG: Value is CAREFUL and logged
                WARNING: Value is WARNING and non logged
                WARNING_LOG: Value is WARNING and logged
                CRITICAL: Value is CRITICAL and non logged
                CRITICAL_LOG: Value is CRITICAL and logged
            optional: True if the stat is optional (display only if space is available)
            additional: True if the stat is additional (display only if space is available after optional)
            spittable: Line can be split to fit on the screen (default is not)
        """
        return {
            'msg': msg,
            'decoration': decoration,
            'optional': optional,
            'additional': additional,
            'splittable': splittable,
        }

    def curse_new_line(self):
        """Go to a new line."""
        return self.curse_add_line('\n')

    def curse_add_stat(self, key, width=None, header='', display_key=True, separator='', trailer=''):
        """Return a list of dict messages with the 'key: value' result

          <=== width ===>
        __key     : 80.5%__
        | |       | |    |_ trailer
        | |       | |_ self.stats[key]
        | |       |_ separator
        | |_ 'short_name' description or key or nothing if display_key is True
        |_ header

        Instead of:
            msg = '  {:8}'.format('idle:')
            ret.append(self.curse_add_line(msg, optional=self.get_views(key='idle', option='optional')))
            msg = '{:5.1f}%'.format(self.stats['idle'])
            ret.append(self.curse_add_line(msg, optional=self.get_views(key='idle', option='optional')))

        Use:
            ret.extend(self.curse_add_stat('idle', width=15, header='  '))

        """
        if key not in self.stats:
            return []

        # Check if a shortname is defined
        if not display_key:
            key_name = ''
        elif key in self.fields_description and 'short_name' in self.fields_description[key]:
            key_name = self.fields_description[key]['short_name']
        else:
            key_name = key

        # Check if unit is defined and get the short unit char in the unit_sort dict
        if (
            key in self.fields_description
            and 'unit' in self.fields_description[key]
            and self.fields_description[key]['unit'] in fields_unit_short
        ):
            # Get the shortname
            unit_short = fields_unit_short[self.fields_description[key]['unit']]
        else:
            unit_short = ''

        # Check if unit is defined and get the unit type unit_type dict
        if (
            key in self.fields_description
            and 'unit' in self.fields_description[key]
            and self.fields_description[key]['unit'] in fields_unit_type
        ):
            # Get the shortname
            unit_type = fields_unit_type[self.fields_description[key]['unit']]
        else:
            unit_type = 'float'

        # Is it a rate ? Yes, get the pre-computed rate value
        if key in self.fields_description and self.fields_description[key].get('rate', False) is True:
            value = self.stats.get(key + '_rate_per_sec', None)
        else:
            value = self.stats.get(key, None)

        if width is None:
            msg_item = header + f'{key_name}' + separator
            msg_template_float = '{:.1f}{}'
            msg_template = '{}{}'
        else:
            # Define the size of the message
            # item will be on the left
            # value will be on the right
            msg_item = header + '{:{width}}'.format(key_name, width=width - 7) + separator
            msg_template_float = '{:5.1f}{}'
            msg_template = '{:>5}{}'

        if value is None:
            msg_value = msg_template.format('-', '')
        elif unit_type == 'float':
            msg_value = msg_template_float.format(value, unit_short)
        elif 'min_symbol' in self.fields_description[key]:
            msg_value = msg_template.format(
                self.auto_unit(int(value), min_symbol=self.fields_description[key]['min_symbol']), unit_short
            )
        else:
            msg_value = msg_template.format(int(value), unit_short)

        # Add the trailer
        msg_value = msg_value + trailer

        decoration = self.get_views(key=key, option='decoration') if value is not None else 'DEFAULT'
        optional = self.get_views(key=key, option='optional')

        return [
            self.curse_add_line(msg_item, optional=optional),
            self.curse_add_line(msg_value, decoration=decoration, optional=optional),
        ]

    @property
    def align(self):
        """Get the curse align."""
        return self._align

    @align.setter
    def align(self, value):
        """Set the curse align.

        value: left, right, bottom.
        """
        self._align = value

    def auto_unit(self, number, low_precision=False, min_symbol='K', none_symbol='-'):
        """Return a nice human-readable string out of number."""
        return auto_unit(number, low_precision=low_precision, min_symbol=min_symbol, none_symbol=none_symbol)

    def trend_msg(self, trend, significant=1):
        """Return the trend message.

        Do not take into account if trend < significant
        """
        ret = '-'
        if trend is None:
            ret = ' '
        elif trend > significant:
            ret = unicode_message('ARROW_UP', self.args)
        elif trend < -significant:
            ret = unicode_message('ARROW_DOWN', self.args)
        return ret

    def _check_decorator(fct):
        """Check decorator for update method.

        It checks:
        - if the plugin is enabled.
        - if the refresh_timer is finished
        """

        def wrapper(self, *args, **kw):
            if self.is_enabled() and (self.refresh_timer.finished() or self.stats == self.get_init_value):
                # Run the method
                ret = fct(self, *args, **kw)
                # Reset the timer
                self.refresh_timer.set(self.get_refresh())
                self.refresh_timer.reset()
            else:
                # No need to call the method
                # Return the last result available
                ret = self.stats
            return ret

        return wrapper

    def _log_result_decorator(fct):
        """Log (DEBUG) the result of the function fct."""

        def wrapper(*args, **kw):
            counter = Counter()
            ret = fct(*args, **kw)
            duration = counter.get()
            class_name = args[0].__class__.__name__
            class_module = args[0].__class__.__module__
            logger.debug(f"{class_name} {class_module} {fct.__name__} return {ret} in {duration} seconds")
            return ret

        return wrapper

    def _manage_rate(fct):
        """Manage rate decorator for update method."""

        def compute_rate(self, stat, stat_previous):
            if stat_previous is None:
                return stat

            # 1) set _gauge for all the rate fields
            # 2) compute the _rate_per_sec
            # 3) set the original field to the delta between the current and the previous value
            for field in self.fields_description:
                # Check if the field exist (avoid error on some OS where some fields are not available)
                if field not in stat:
                    continue
                # For all the field with the rate=True flag
                # if 'rate' in self.fields_description[field] and self.fields_description[field]['rate'] is True:
                if self.fields_description[field].get('rate', False):
                    # Create a new metadata with the gauge
                    stat['time_since_update'] = self.time_since_last_update
                    stat[field + '_gauge'] = stat[field]
                    if field + '_gauge' in stat_previous and stat[field] and stat_previous[field + '_gauge']:
                        # The stat becomes the delta between the current and the previous value
                        stat[field] = stat[field] - stat_previous[field + '_gauge']
                        # Compute the rate
                        if self.time_since_last_update > 0:
                            stat[field + '_rate_per_sec'] = stat[field] // self.time_since_last_update
                        else:
                            stat[field] = 0
                            stat[field + '_rate_per_sec'] = 0
                    else:
                        # Avoid strange rate at the first run
                        stat[field] = 0
                        stat[field + '_rate_per_sec'] = 0
            return stat

        def compute_rate_on_list(self, stats, stats_previous):
            if stats_previous is None:
                return stats

            for stat in stats:
                olds = [i for i in stats_previous if i[self.get_key()] == stat[self.get_key()]]
                if len(olds) == 1:
                    compute_rate(self, stat, olds[0])
            return stats

        def wrapper(self, *args, **kw):
            # Call the father method
            stats = fct(self, *args, **kw)

            # Get the time since the last update
            self.time_since_last_update = getTimeSinceLastUpdate(self.plugin_name)

            # Compute the rate
            if isinstance(stats, dict):
                # Stats is a dict
                compute_rate(self, stats, self.stats_previous)
            elif isinstance(stats, list):
                # Stats is a list
                compute_rate_on_list(self, stats, self.stats_previous)

            # Memorized the current stats for next run
            self.stats_previous = copy.deepcopy(stats)

            return stats

        return wrapper

    def _manage_mmm(fct):
        """Manage MMM (Min/Max/Mean) decorator for update method.

        Automatically computes and adds min/max/mean fields for any field with mmm=True.
        """

        def wrapper(self, *args, **kw):
            # Call the father method
            stats = fct(self, *args, **kw)

            # Update MMM fields
            if isinstance(stats, dict):
                # Stats is a dict
                self._update_mmm_fields(stats)
            elif isinstance(stats, list):
                # Stats is a list
                self._update_mmm_fields_on_list(stats)

            return stats

        return wrapper

    # Mandatory to call the decorator in child classes
    _check_decorator = staticmethod(_check_decorator)
    _log_result_decorator = staticmethod(_log_result_decorator)
    _manage_rate = staticmethod(_manage_rate)
    _manage_mmm = staticmethod(_manage_mmm)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""CPU percent stats shared between CPU and Quicklook plugins."""

import platform
from typing import TypedDict

import psutil

from glances.logger import logger
from glances.timer import Timer

__all__ = ["cpu_percent"]

CPU_IMPLEMENTERS = {
    0x41: 'ARM Limited',
    0x42: 'Broadcom',
    0x43: 'Cavium',
    0x44: 'DEC',
    0x46: 'Fujitsu',
    0x48: 'HiSilicon',
    0x49: 'Infineon Technologies',
    0x4D: 'Motorola/Freescale',
    0x4E: 'NVIDIA',
    0x50: 'Applied Micro (APM)',
    0x51: 'Qualcomm',
    0x53: 'Samsung',
    0x56: 'Marvell',
    0x61: 'Apple',
    0x66: 'Faraday',
    0x69: 'Intel',
    0x6D: 'Microsoft',
    0x70: 'Phytium',
    0xC0: 'Ampere Computing',
}

CPU_PARTS = {
    # ARM Limited (0x41)
    0x41: {
        0xD03: 'Cortex-A53',
        0xD04: 'Cortex-A35',
        0xD05: 'Cortex-A55',
        0xD06: 'Cortex-A65',
        0xD07: 'Cortex-A57',
        0xD08: 'Cortex-A72',
        0xD09: 'Cortex-A73',
        0xD0A: 'Cortex-A75',
        0xD0B: 'Cortex-A76',
        0xD0C: 'Neoverse N1',
        0xD0D: 'Cortex-A77',
        0xD0E: 'Cortex-A76AE',
        0xD13: 'Cortex-R52',
        0xD20: 'Cortex-M23',
        0xD21: 'Cortex-M33',
        0xD40: 'Neoverse V1',
        0xD41: 'Cortex-A78',
        0xD42: 'Cortex-A78AE',
        0xD43: 'Cortex-A65AE',
        0xD44: 'Cortex-X1',
        0xD46: 'Cortex-A510',
        0xD47: 'Cortex-A710',
        0xD48: 'Cortex-X2',
        0xD49: 'Neoverse N2',
        0xD4A: 'Neoverse E1',
        0xD4B: 'Cortex-A78C',
        0xD4C: 'Cortex-X1C',
        0xD4D: 'Cortex-A715',
        0xD4E: 'Cortex-X3',
        0xD4F: 'Neoverse V2',
        0xD80: 'Cortex-A520',
        0xD81: 'Cortex-A720',
        0xD82: 'Cortex-X4',
        0xD84: 'Neoverse V3',
        0xD85: 'Cortex-X925',
        0xD87: 'Cortex-A725',
    },
    # Apple (0x61)
    0x61: {
        0x000: 'Swift',
        0x001: 'Cyclone',
        0x002: 'Typhoon',
        0x003: 'Twister',
        0x004: 'Hurricane',
        0x005: 'Monsoon/Mistral',
        0x006: 'Vortex/Tempest',
        0x007: 'Lightning/Thunder',
        0x008: 'Firestorm/Icestorm (M1)',
        0x009: 'Avalanche/Blizzard (M2)',
        0x00E: 'Everest/Sawtooth (M3)',
        0x010: 'Blizzard/Avalanche (A16)',
        0x011: 'Coll (M4)',
    },
    # Qualcomm (0x51)
    0x51: {
        0x00F: 'Scorpion',
        0x02D: 'Scorpion',
        0x04D: 'Krait',
        0x06F: 'Krait',
        0x201: 'Kryo',
        0x205: 'Kryo',
        0x211: 'Kryo',
        0x800: 'Kryo 260/280 Gold (Cortex-A73)',
        0x801: 'Kryo 260/280 Silver (Cortex-A53)',
        0x802: 'Kryo 385 Gold (Cortex-A75)',
        0x803: 'Kryo 385 Silver (Cortex-A55)',
        0x804: 'Kryo 485 Gold (Cortex-A76)',
        0x805: 'Kryo 485 Silver (Cortex-A55)',
        0xC00: 'Falkor',
        0xC01: 'Saphira',
    },
    # Samsung (0x53)
    0x53: {
        0x001: 'Exynos M1/M2',
        0x002: 'Exynos M3',
        0x003: 'Exynos M4',
        0x004: 'Exynos M5',
    },
    # NVIDIA (0x4e)
    0x4E: {
        0x000: 'Denver',
        0x003: 'Denver 2',
        0x004: 'Carmel',
    },
    # Marvell (0x56)
    0x56: {
        0x131: 'Feroceon 88FR131',
        0x581: 'PJ4/PJ4b',
        0x584: 'PJ4B-MP',
    },
    # Cavium (0x43)
    0x43: {
        0x0A0: 'ThunderX',
        0x0A1: 'ThunderX 88XX',
        0x0A2: 'ThunderX 81XX',
        0x0A3: 'ThunderX 83XX',
        0x0AF: 'ThunderX2 99xx',
        0x0B0: 'OcteonTX2',
        0x0B1: 'OcteonTX2 T98',
        0x0B2: 'OcteonTX2 T96',
        0x0B3: 'OcteonTX2 F95',
        0x0B4: 'OcteonTX2 F95N',
        0x0B5: 'OcteonTX2 F95MM',
    },
    # Broadcom (0x42)
    0x42: {
        0x00F: 'Brahma B15',
        0x100: 'Brahma B53',
        0x516: 'Vulcan',
    },
    # HiSilicon (0x48)
    0x48: {
        0xD01: 'Kunpeng-920',
        0xD40: 'Cortex-A76 (Kirin)',
    },
    # Ampere (0xc0)
    0xC0: {
        0xAC3: 'Ampere-1',
        0xAC4: 'Ampere-1a',
    },
    # Fujitsu (0x46)
    0x46: {
        0x001: 'A64FX',
    },
    # Intel (0x69) - ARM-based chips
    0x69: {
        0x200: 'i80200',
        0x210: 'PXA250A',
        0x212: 'PXA210A',
        0x242: 'i80321-400',
        0x243: 'i80321-600',
        0x290: 'PXA250B/PXA26x',
        0x292: 'PXA210B',
        0x2C2: 'i80321-400-B0',
        0x2C3: 'i80321-600-B0',
        0x2D0: 'PXA250C/PXA255/PXA26x',
        0x2D2: 'PXA210C',
        0x411: 'PXA27x',
        0x41C: 'IPX425-533',
        0x41D: 'IPX425-400',
        0x41F: 'IPX425-266',
        0x682: 'PXA32x',
        0x683: 'PXA930/PXA935',
        0x688: 'PXA30x',
        0x689: 'PXA31x',
    },
}


class CpuInfo(TypedDict):
    cpu_name: str
    cpu_hz: float | None
    cpu_hz_current: float | None


class PerCpuPercentInfo(TypedDict):
    key: str
    cpu_number: int
    total: float
    user: float
    system: float
    idle: float
    nice: float | None
    iowait: float | None
    irq: float | None
    softirq: float | None
    steal: float | None
    guest: float | None
    guest_nice: float | None
    dpc: float | None
    interrupt: float | None


class CpuPercent:
    """Get and store the CPU percent."""

    def __init__(self, cached_timer_cpu: int = 2):
        # cached_timer_cpu is the minimum time interval between stats updates
        # since last update is passed (will retrieve old cached info instead)
        self.cached_timer_cpu = cached_timer_cpu
        # psutil.cpu_freq() consumes lots of CPU
        # So refresh CPU frequency stats every refresh * 2
        self.cached_timer_cpu_info = cached_timer_cpu * 2

        # Get CPU name
        self.timer_cpu_info = Timer(0)
        self.cpu_info: CpuInfo = {'cpu_name': self.__get_cpu_name(), 'cpu_hz_current': None, 'cpu_hz': None}

        # Warning from PsUtil documentation
        # The first time this function is called with interval = 0.0 or None
        # it will return a meaningless 0.0 value which you are supposed to ignore.
        self.timer_cpu = Timer(0)
        self.cpu_percent = self._compute_cpu()
        self.timer_percpu = Timer(0)
        self.percpu_percent = self._compute_percpu()

    def get_key(self):
        """Return the key of the per CPU list."""
        return 'cpu_number'

    def get_info(self) -> CpuInfo:
        """Get additional information about the CPU"""
        # Never update more than 1 time per cached_timer_cpu_info
        if self.timer_cpu_info.finished() and hasattr(psutil, 'cpu_freq'):
            # Get the CPU freq current/max
            try:
                cpu_freq = psutil.cpu_freq()
            except Exception as e:
                logger.debug(f'Can not grab CPU information ({e})')
            else:
                if hasattr(cpu_freq, 'current'):
                    self.cpu_info['cpu_hz_current'] = cpu_freq.current
                else:
                    self.cpu_info['cpu_hz_current'] = None
                if hasattr(cpu_freq, 'max') and cpu_freq.max != 0.0:
                    self.cpu_info['cpu_hz'] = cpu_freq.max
                else:
                    self.cpu_info['cpu_hz'] = None
                # Reset timer for cache
                self.timer_cpu_info.reset(duration=self.cached_timer_cpu_info)
        return self.cpu_info

    @staticmethod
    def __get_cpu_name() -> str:
        # Get the CPU name once from the /proc/cpuinfo file
        # Read the first line with the "model name" ("Model" for Raspberry Pi)
        ret = f'CPU {platform.processor()}'
        try:
            cpuinfo_lines = open('/proc/cpuinfo').readlines()
        except (FileNotFoundError, PermissionError):
            logger.debug("No permission to read '/proc/cpuinfo'")
            return ret

        cpu_implementer = None
        for line in cpuinfo_lines:
            # Look for the CPU name
            if line.startswith('model name') or line.startswith('Model') or line.startswith('cpu model'):
                return line.split(':')[1].strip()
            # Look for the CPU name on ARM architecture (see #3127)
            if line.startswith('CPU implementer'):
                cpu_implementer = CPU_IMPLEMENTERS.get(int(line.split(':')[1].strip(), 16), ret)
                ret = cpu_implementer
            if line.startswith('CPU part') and cpu_implementer in CPU_PARTS:
                cpu_part = CPU_PARTS[cpu_implementer].get(int(line.split(':')[1].strip(), 16), 'Unknown')
                ret = f'{cpu_implementer} {cpu_part}'

        return ret

    def get_cpu(self) -> float:
        """Update and/or return the CPU using the psutil library."""
        # Never update more than 1 time per cached_timer_cpu
        if self.timer_cpu.finished():
            # Reset timer for cache
            self.timer_cpu.reset(duration=self.cached_timer_cpu)
            # Update the stats
            self.cpu_percent = self._compute_cpu()
        return self.cpu_percent

    @staticmethod
    def _compute_cpu() -> float:
        return psutil.cpu_percent(interval=0.0)

    def get_percpu(self) -> list[PerCpuPercentInfo]:
        """Update and/or return the per CPU list using the psutil library."""
        # Never update more than 1 time per cached_timer_cpu
        if self.timer_percpu.finished():
            # Reset timer for cache
            self.timer_percpu.reset(duration=self.cached_timer_cpu)
            # Update stats
            self.percpu_percent = self._compute_percpu()
        return self.percpu_percent

    def _compute_percpu(self) -> list[PerCpuPercentInfo]:
        psutil_percpu = enumerate(psutil.cpu_times_percent(interval=0.0, percpu=True))
        return [
            {
                'key': self.get_key(),
                'cpu_number': cpu_number,
                'total': round(100 - cpu_times.idle, 1),
                'user': cpu_times.user,
                'system': cpu_times.system,
                'idle': cpu_times.idle,
                'nice': cpu_times.nice if hasattr(cpu_times, 'nice') else None,
                'iowait': cpu_times.iowait if hasattr(cpu_times, 'iowait') else None,
                'irq': cpu_times.irq if hasattr(cpu_times, 'irq') else None,
                'softirq': cpu_times.softirq if hasattr(cpu_times, 'softirq') else None,
                'steal': cpu_times.steal if hasattr(cpu_times, 'steal') else None,
                'guest': cpu_times.guest if hasattr(cpu_times, 'guest') else None,
                'guest_nice': cpu_times.steal if hasattr(cpu_times, 'guest_nice') else None,
                'dpc': cpu_times.dpc if hasattr(cpu_times, 'dpc') else None,
                'interrupt': cpu_times.interrupt if hasattr(cpu_times, 'interrupt') else None,
            }
            for cpu_number, cpu_times in psutil_percpu
        ]


# CpuPercent instance shared between plugins
cpu_percent = CpuPercent()



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Intel Extension unit for Glances' GPU plugin.

The class grabs the stats from the /sys/class/drm/ directory.

See: https://github.com/nicolargo/glances/issues/994
"""

# Example
# /sys/class/drm/card0
# ├── gt_act_freq_mhz
# ├── gt_boost_freq_mhz
# ├── gt_cur_freq_mhz
# ├── gt_max_freq_mhz
# ├── gt_min_freq_mhz
# ├── gt_RP0_freq_mhz
# ├── gt_RP1_freq_mhz
# ├── gt_RPn_freq_mhz

import functools
import glob
import os
import re

DRM_ROOT_FOLDER: str = '/sys/class/drm'
DEVICE_FOLDER_PATTERN: str = 'card[0-9]'
INTELGPU_IDS_FILE: str = '/usr/share/misc/pci.ids'
PCI_DEVICE_VENDOR: str = 'device/vendor'
PCI_DEVICE_ID: str = 'device/device'
PCI_ACT_FRQ_MHZ: str = 'gt_act_freq_mhz'
PCI_MAX_FRQ_MHZ: str = 'gt_max_freq_mhz'


class IntelGPU:
    """Intel GPU card class."""

    def __init__(self, drm_root_folder: str = DRM_ROOT_FOLDER):
        """Init Intel GPU card class."""
        self.drm_root_folder = drm_root_folder
        self.device_folders = get_device_list(drm_root_folder)

    def exit(self):
        """Close Intel GPU class."""

    def get_device_stats(self):
        """Get Intel GPU stats."""
        stats = []

        for index, device in enumerate(self.device_folders):
            device_stats = {}
            # Dictionary key is the GPU_ID
            device_stats['key'] = 'gpu_id'
            # GPU id (for multiple GPU, start at 0)
            device_stats['gpu_id'] = f'intel{index}'
            # GPU name
            device_stats['name'] = get_device_name(device)
            # Memory consumption in % (not available on all GPU)
            device_stats['mem'] = get_mem(device)
            # Processor consumption in % of frequency
            device_stats['proc'] = get_proc(device)
            # Processor temperature in °C
            device_stats['temperature'] = get_temperature(device)
            # Fan speed in %
            device_stats['fan_speed'] = get_fan_speed(device)
            stats.append(device_stats)

        return stats


def get_device_list(drm_root_folder: str) -> list[str]:
    """Return a list of path to the device stats."""
    ret = []
    for device_folder in glob.glob(DEVICE_FOLDER_PATTERN, root_dir=drm_root_folder):
        if os.path.isfile(os.path.join(drm_root_folder, device_folder, PCI_ACT_FRQ_MHZ)):
            # If the GPU act freq file is present then take the card into account
            ret.append(os.path.join(drm_root_folder, device_folder))
    return ret


def read_file(*path_segments: str) -> str | None:
    """Return content of file or None if not accessible."""
    path = os.path.join(*path_segments)
    if os.path.isfile(path):
        try:
            with open(path) as f:
                return f.read().strip()
        except (PermissionError, OSError):
            # File exists but is not readable (e.g. Snap strict confinement)
            # Graceful degradation: caller will use a default value instead
            return None
    return None


@functools.cache
def get_device_name(device_folder: str) -> str:
    """Return the GPU name."""
    device_vendor = read_file(device_folder, PCI_DEVICE_VENDOR)
    device_id = read_file(device_folder, PCI_DEVICE_ID)
    intelgpu_ids = read_file(INTELGPU_IDS_FILE)

    if device_vendor and device_id and intelgpu_ids:
        device_vendor = device_vendor.strip()[2:].lower()
        device_id = device_id.strip()[2:].lower()

        pattern = rf'^{device_vendor}[ \t]+.+\n(?:(?![0-9a-f]{{4}}).+\n)*?\t{device_id}[ \t]+(.+)$'
        match = re.search(pattern, intelgpu_ids, re.MULTILINE)
        if match:
            name = match.group(1)
            bracket_match = re.search(r'\[(.+)\]', name)
            return bracket_match.group(1) if bracket_match else name

    return 'Intel GPU'


def get_mem(device_folder: str) -> int | None:
    """Return the memory consumption in %."""
    return None


def get_proc(device_folder: str) -> int | None:
    """Return the processor consumption in %."""
    act_freq = read_file(device_folder, PCI_ACT_FRQ_MHZ)
    max_freq = read_file(device_folder, PCI_MAX_FRQ_MHZ)
    if act_freq and max_freq:
        return round(int(act_freq) / int(max_freq) * 100)
    return None


def get_temperature(device_folder: str) -> int | None:
    """Return the processor temperature in °C (mean of all HWMON)"""
    return None


def get_fan_speed(device_folder: str) -> int | None:
    """Return the fan speed in %."""
    return None



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Load plugin."""

import os

import psutil

from glances.logger import logger
from glances.plugins.core import CorePlugin
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
fields_description = {
    'min1': {
        'description': 'Average sum of the number of processes \
waiting in the run-queue plus the number currently executing \
over 1 minute.',
        'unit': 'float',
        'mmm': True,
    },
    'min5': {
        'description': 'Average sum of the number of processes \
waiting in the run-queue plus the number currently executing \
over 5 minutes.',
        'unit': 'float',
    },
    'min15': {
        'description': 'Average sum of the number of processes \
waiting in the run-queue plus the number currently executing \
over 15 minutes.',
        'unit': 'float',
    },
    'cpucore': {'description': 'Total number of CPU core.', 'unit': 'number'},
}

# SNMP OID
# 1 minute Load: .1.3.6.1.4.1.2021.10.1.3.1
# 5 minute Load: .1.3.6.1.4.1.2021.10.1.3.2
# 15 minute Load: .1.3.6.1.4.1.2021.10.1.3.3
snmp_oid = {
    'min1': '1.3.6.1.4.1.2021.10.1.3.1',
    'min5': '1.3.6.1.4.1.2021.10.1.3.2',
    'min15': '1.3.6.1.4.1.2021.10.1.3.3',
}

# Define the history items list
# All items in this list will be historised if the --enable-history tag is set
items_history_list = [
    {'name': 'min1', 'description': '1 minute load'},
    {'name': 'min5', 'description': '5 minutes load'},
    {'name': 'min15', 'description': '15 minutes load'},
]

# Get the number of logical CPU core only once
# the variable is also shared with the QuickLook plugin
nb_log_core = 1
nb_phys_core = 1
try:
    core = CorePlugin().update()
except Exception as e:
    logger.warning(f'Error: Can not retrieve the CPU core number (set it to 1) ({e})')
else:
    if 'log' in core:
        nb_log_core = core['log']
    if 'phys' in core:
        nb_phys_core = core['phys']


class LoadPlugin(GlancesPluginModel):
    """Glances load plugin.

    stats is a dict
    """

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(
            args=args, config=config, items_history_list=items_history_list, fields_description=fields_description
        )

        # We want to display the stat in the curse interface
        self.display_curse = True

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update load stats."""
        # Init new stats
        stats = self.get_init_value()

        if self.input_method == 'local':
            # Update stats using the standard system lib

            # Get the load using the os standard lib
            load = load_average()
            if load is None:
                stats = self.get_init_value()
            else:
                stats = {'min1': load[0], 'min5': load[1], 'min15': load[2], 'cpucore': log_core()}

        elif self.input_method == 'snmp':
            # Update stats using SNMP
            stats = self.get_stats_snmp(snmp_oid=snmp_oid)

            if stats['min1'] == '':
                return self.get_init_value()

            # Python 3 return a dict like:
            # {'min1': "b'0.08'", 'min5': "b'0.12'", 'min15': "b'0.15'"}
            for k, v in stats.items():
                stats[k] = float(v)

            stats['cpucore'] = log_core()

        # Update the stats
        self.stats = stats

        return self.stats

    def update_views(self):
        """Update stats views."""
        # Call the father's method
        super().update_views()

        # Add specifics information
        try:
            # Alert and log
            self.views['min15']['decoration'] = self.get_alert_log(
                self.stats['min15'], maximum=100 * self.stats['cpucore']
            )
            # Alert only
            self.views['min5']['decoration'] = self.get_alert(self.stats['min5'], maximum=100 * self.stats['cpucore'])
        except KeyError:
            # try/except mandatory for Windows compatibility (no load stats)
            pass

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        # Init the return message
        ret = []

        # Only process if stats exist, not empty (issue #871) and plugin not disabled
        if not self.stats or (self.stats == {}) or self.is_disabled():
            return ret

        # Build the string message
        # Header
        msg = '{:4}'.format('LOAD')
        ret.append(self.curse_add_line(msg, "TITLE"))
        msg = ' {:1}'.format(self.trend_msg(self.get_trend('min1')))
        ret.append(self.curse_add_line(msg))
        # Core number
        if 'cpucore' in self.stats and self.stats['cpucore'] > 0:
            msg = '{:3}core'.format(int(self.stats['cpucore']))
            ret.append(self.curse_add_line(msg))
        # Loop over 1min, 5min and 15min load
        for load_time in ['1', '5', '15']:
            ret.append(self.curse_new_line())
            msg = '{:7}'.format(f'{load_time} min')
            ret.append(self.curse_add_line(msg))
            if args and args.disable_irix and log_core() != 0:
                # Enable Irix mode for load (see issue #1554)
                load_stat = self.stats[f'min{load_time}'] / log_core() * 100
                msg = f'{load_stat:>5.1f}%'
            else:
                # Default mode for load
                load_stat = self.stats[f'min{load_time}']
                msg = f'{load_stat:>6.2f}'
            ret.append(self.curse_add_line(msg, self.get_views(key=f'min{load_time}', option='decoration')))

        return ret


def log_core():
    """Get the number of logical CPU core."""
    return nb_log_core


def phys_core():
    """Get the number of physical CPU core."""
    return nb_phys_core


def load_average(percent: bool = False):
    """Get load average. On both Linux and Windows thanks to PsUtil

    if percent is True, return the load average in percent
    Ex: if you only have one CPU core and the load average is 1.0, then return 100%"""
    ret = None
    try:
        ret = psutil.getloadavg()
    except (AttributeError, OSError):
        try:
            ret = os.getloadavg()
        except (AttributeError, OSError):
            pass

    if ret and percent:
        return tuple([round(i / log_core() * 100, 1) for i in ret])
    return ret
